# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# © 2024–2025 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Markdown to HTML conversion."""

import argparse
import sys
from collections.abc import Collection

import yaml
from markdown_it import MarkdownIt
from mdit_py_plugins.anchors import anchors_plugin
from mdit_py_plugins.front_matter import front_matter_plugin
from path import Path

from stepup.core.api import amend, getenv
from stepup.core.render_jinja import render_jinja

__all__ = ("convert_markdown",)


def main(argv: list[str] | None = None):
    """Main program."""
    args = parse_args(argv)
    if not args.markdown.endswith(".md"):
        raise ValueError("The markdown file must end with the .md extension.")
    if len(args.css) == 0:
        args.css = getenv("REPREP_MARKDOWN_CSS", multi=True, back=True)
    with open(args.markdown) as fm, open(args.html, "w") as fh:
        fh.write(convert_markdown(fm.read(), args.css, args.html.parent))


def parse_args(argv: list[str] | None = None) -> argparse.Namespace:
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(
        prog="rr-convert-markdown", description="Convert Markdown to HTML"
    )
    parser.add_argument("markdown", type=Path, help="A Markdown file with extension `.md`")
    parser.add_argument("html", type=Path, help="A HTML output filename")
    parser.add_argument(
        "--css",
        type=Path,
        nargs="+",
        default=(),
        help="Local CSS files to link to in the HTML header. "
        "The default value is ${REPREP_MARKDOWN_CSS} (if defined) "
        "and it is interpreted as a colon-separated list.",
    )
    return parser.parse_args(argv)


HTML_TEMPLATE = """\
<!DOCTYPE html>
<html>
<head>
  <meta charset="utf-8" />
  <meta name="viewport" content="width=device-width, initial-scale=1.0, user-scalable=yes" />
  <title>{{ title }}</title>
  {{ css | indent(width=2) }}
</head>
<body>
  {{ body | indent(width=2) }}
</body>
</html>
"""


def convert_markdown(
    text_md: str,
    paths_css: str | Collection[str] = (),
    parent_html: str = "./",
) -> str:
    """Convert Markdown to HTML.

    Parameters
    ----------
    text_md
        The markdown source text.
    paths_css
        Path of a local CSS file, or a list of multiple such paths,
        to be included in the HTML header.
        Note that one may also specify CSS file in the markdown header.
    parent_html
        The parent path of the HTML file.
        The CSS paths are rewritten to become relative to this parent path.

    Returns
    -------
    html
        The HTML conversion.
    """
    md = MarkdownIt().use(front_matter_plugin).use(anchors_plugin)

    # Convert to HTML and get front matter.
    tokens = md.parse(text_md)
    body = md.renderer.render(tokens, md.options, {})
    meta = {}
    if tokens[0].type == "front_matter":
        meta = yaml.safe_load(tokens[0].content)

    # Collect CSS paths from the source and amend them
    parent_html = Path(parent_html)
    paths_doc_css = meta.get("css", [])
    if isinstance(paths_doc_css, str):
        paths_doc_css = [paths_doc_css]
    amend(inp=[parent_html / path_css for path_css in paths_doc_css if "://" not in path_css])

    # Convert the given CSS files to be relative to the parent of the HTML file.
    if isinstance(paths_css, str):
        paths_css = paths_css.split(":")
    paths_css = [
        Path(path_css).relpath(parent_html) for path_css in paths_css if "://" not in path_css
    ]
    paths_css.extend(paths_doc_css)

    # Use Jinja to finalize the HTML page.
    variables = {
        "body": body,
        "title": meta.get("title", "Untitled"),
        "css": "\n".join(f'<link rel="stylesheet" href="{path_css}" />' for path_css in paths_css),
    }
    return render_jinja("HTML_TEMPLATE", variables, str_in=HTML_TEMPLATE)


if __name__ == "__main__":
    main(sys.argv[1:])



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# © 2024–2025 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
r"""Identification of dependencies from LaTeX sources."""

import re

from path import Path

from stepup.core.api import amend
from stepup.core.utils import filter_dependencies

RE_OPTIONS = re.MULTILINE | re.DOTALL
RE_INPUT = re.compile(r"\\input\s*\{([^}]*)}", RE_OPTIONS)
RE_VERBATIMINPUT = re.compile(r"\\verbatiminput\s*\{([^}]*)}", RE_OPTIONS)
RE_INCLUDEGRAPHICS = re.compile(r"\\includegraphics(?:\s*\[[^]]*])?\s*\{([^}]*)}", RE_OPTIONS)
RE_BIBLIOGRAPHY = re.compile(r"\\bibliography\s*\{([^}]*)}", RE_OPTIONS)
RE_IMPORT = re.compile(r"\\import\s*\{([^}]*)}\s*\{([^}]*)}", RE_OPTIONS)
RE_INCLUDPDF = re.compile(r"\\includepdf(?:\s*\[[^]]*])?\s*\{([^}]*)}", RE_OPTIONS)


def cleanup_path(path, ext=None):
    """Clean up a path of a dependency extracted from a LaTeX source.

    Parameters
    ----------
    path
        The path derived from the LaTeX source.
    ext
        The extension to be added

    Returns
    -------
    path_clean
        A cleaned up file name, including the path of the dirname.
    """
    path = path.replace("{", "")
    path = path.replace("}", "")
    path = re.sub(r"\s+", " ", path)
    path = Path(path.strip())
    if "." not in path.basename() and ext is not None:
        path = path.with_suffix(ext)
    return path.normpath()


def iter_latex_references(tex_no_comments):
    r"""Loop over file references in a TeX source without comments.

    Parameters
    ----------
    tex_no_comments
        The contents of a TeX source files from which comments were stripped.

    Yields
    ------
    relative_path
        The change in current directory implied by a TeX command.
        (This is only relevant for \import, no change in directory otherwise.)
    filename
        The filename of the file included, may include directory.
    ext
        The extension one may add if not given.
        (Approximate guess, because the correct extension for figures
        depends on details of the LaTeX compiler.)
    """
    for fn_inc in re.findall(RE_INPUT, tex_no_comments):
        yield ".", fn_inc, ".tex"
    for fn_inc in re.findall(RE_VERBATIMINPUT, tex_no_comments):
        yield ".", fn_inc, ".txt"
    for fn_inc in re.findall(RE_INCLUDEGRAPHICS, tex_no_comments):
        yield ".", fn_inc, ".pdf"
    for fn_inc in re.findall(RE_BIBLIOGRAPHY, tex_no_comments):
        yield ".", fn_inc, ".bib"
    for new_root, fn_inc in re.findall(RE_IMPORT, tex_no_comments):
        yield new_root, fn_inc, ".tex"
    for fn_inc in re.findall(RE_INCLUDPDF, tex_no_comments):
        yield ".", fn_inc, ".pdf"


def scan_latex_deps(path_tex, tex_root=None, do_amend=True):
    """Scan LaTeX source code for dependencies.

    Parameters
    ----------
    path_tex
        The path to the LaTeX source to scan.
    tex_root
        The directory with respect to which the latex file references should be interpreted.
    do_amend
        When True, all opened TeX files are amended as input to the current step.

    Returns
    -------
    inp
        Filenames to be added as amended inputs.
    bib
        BibTeX files, a special case of inputs to amend.
    out
        Filenames to be added as amended outputs.
    vol
        Filenames to be added as amended volatile outputs.
    """
    inp = set()
    out = set()
    vol = set()
    bib = set()

    path_tex = Path(path_tex)
    if path_tex.is_file():
        tex_root = path_tex.parent.normpath() if tex_root is None else Path(tex_root)
        with open(path_tex) as fh:
            stripped = []
            for line in fh:
                if "%REPREP ignore" in line:
                    pass
                elif line.startswith("%REPREP inp "):
                    inp.add((tex_root / line[12:].strip()).normpath())
                elif line.startswith("%REPREP out "):
                    out.add((tex_root / line[12:].strip()).normpath())
                elif line.startswith("%REPREP vol "):
                    vol.add((tex_root / line[12:].strip()).normpath())
                else:
                    stripped.append(line[: line.find("%")].rstrip())

            # Process the file references
            for new_root, fn_inc, ext in iter_latex_references("\n".join(stripped)):
                new_root = (tex_root / cleanup_path(new_root)).normpath()
                path_inc = (new_root / cleanup_path(fn_inc, ext)).normpath()
                if ext == ".bib":
                    bib.add(path_inc)
                else:
                    inp.add(path_inc)
                if ext == ".tex":
                    sub_inp, sub_bib, sub_out, sub_vol = scan_latex_deps(path_inc, new_root)
                    inp.update(sub_inp)
                    bib.update(sub_bib)
                    out.update(sub_out)
                    vol.update(sub_vol)

    # Filter dependencies to exclude global files
    inp = filter_dependencies(inp)
    bib = filter_dependencies(bib)
    if do_amend:
        amend(inp=[path for path in inp if path.endswith(".tex")])

    return sorted(inp), sorted(bib), sorted(out), sorted(vol)



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# © 2024–2025 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Checking of inventory files."""

import argparse
import sys
from collections.abc import Iterator

from path import Path

from .inventory import FileSummary, check_summary, get_summary, parse_summary

__all__ = ("check_inventory", "iter_inventory", "main")


def main(argv: list[str] | None = None):
    """Main program."""
    parser = argparse.ArgumentParser(
        prog="rr-check-inventory", description="Check an inventory.txt."
    )
    add_parser_args(parser)
    args = parser.parse_args(argv)
    if not args.inventory_txt.endswith(".txt"):
        raise ValueError("The inventory file must end with .txt")
    check_inventory(args.inventory_txt)


def add_parser_args(parser: argparse.ArgumentParser):
    """Define command-line arguments."""
    parser.add_argument("inventory_txt", help="An inventory.txt file generated with RepRep")


def check_subcommand(subparser: argparse.ArgumentParser) -> callable:
    """Add subcommand to check inventory."""
    parser = subparser.add_parser(
        "check-inventory",
        help="Check the inventory file.",
    )
    add_parser_args(parser)
    return check_tool


def check_tool(args: argparse.Namespace) -> int:
    """Check the inventory file."""
    check_inventory(args.inventory_txt)
    return 0


def iter_inventory(path_inventory: str) -> Iterator[FileSummary]:
    with open(path_inventory) as fh:
        for iline, line in enumerate(fh):
            try:
                fs = parse_summary(line)
            except Exception as exc:
                raise ValueError(f"Could not parse line {iline} of {path_inventory}") from exc
            yield fs


def check_inventory(path_inventory: str):
    path_inventory = Path(path_inventory)
    root = path_inventory.parent
    for ref in iter_inventory(path_inventory):
        new = get_summary(root / ref.path, root)
        check_summary(new, ref)


if __name__ == "__main__":
    sys.exit(main(sys.argv))



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# © 2024–2025 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""RepRep Wrapper for the Tectonic LaTeX compiler.

The Tectonic command-line interface deviates quite a bit from the usual LaTeX compilers,
so it is easier to have a separate wrapper for it.
"""

import argparse
import contextlib
import re
import shlex
import sys

from path import Path, TempDir

from stepup.core.api import amend, getenv
from stepup.core.utils import filter_dependencies
from stepup.core.worker import WorkThread

from .make_inventory import write_inventory


def main(argv: list[str] | None = None, work_thread: WorkThread | None = None) -> None:
    """Main program."""
    args = parse_args(argv)
    if work_thread is None:
        work_thread = WorkThread("stub")

    workdir, fn_tex = args.path_tex.splitpath()
    workdir = workdir.normpath()
    if not fn_tex.endswith(".tex"):
        raise ValueError("The LaTeX source must have extension .tex")

    # Get the Tectonic executable
    if args.tectonic is None:
        args.tectonic = getenv("REPREP_TECTONIC", "tectonic")

    # Prepare the command to run Tectonic
    tectonic_args = [args.tectonic, "-c", "minimal", args.path_tex]
    if len(args.tectonic_args) == 0:
        args.tectonic_args = shlex.split(getenv("REPREP_TECTONIC_ARGS", ""))
    tectonic_args.extend(args.tectonic_args)

    with contextlib.ExitStack() as stack:
        if args.keep_deps:
            # Remove any existing make-deps output from a previous run.
            path_dep = Path(args.path_tex.with_suffix(".dep"))
            path_dep.remove_p()
        else:
            # Use a temporary file for the make-deps output.
            path_dep = stack.enter_context(TempDir()) / "tectonic.dep"
        tectonic_args.extend(["--makefile-rules", path_dep])

        # Run Tectonic in the directory of the tex file
        with contextlib.chdir(workdir):
            returncode, stdout, stderr = work_thread.runsh(shlex.join(tectonic_args))
        print(stdout)
        # Get existing input files from the dependency file and amend.
        # Note that the deps file does not escape colons in paths,
        # so the code below assumes one never uses colons in paths.
        inp_paths = []
        out_paths = []
        if path_dep.is_file():
            with open(path_dep) as fh:
                dep_out, dep_inp = fh.read().replace("\\\n", " ").split(":", 1)
                out_paths.extend(workdir / path for path in shlex.split(dep_out))
                inp_paths.extend(workdir / path for path in shlex.split(dep_inp))
        else:
            print(f"Dependency file not created: {path_dep}.", file=sys.stderr)

    # Look for missing input files in the standard error stream and amend them.
    if returncode != 0:
        inp_paths.extend(
            workdir / m.group(1)
            for m in re.finditer(r"`([^`]+)' not found", stderr, flags=re.MULTILINE)
        )
    sys.stderr.write(stderr)
    inp_paths = filter_dependencies(inp_paths)
    amend(inp=inp_paths)

    # Write inventory
    if args.inventory is not None:
        inventory_paths = sorted(inp_paths) + out_paths
        write_inventory(args.inventory, inventory_paths, do_amend=False)

    if returncode != 0:
        # Only use sys.exit in cases of an error,
        # so other programs may call this function without exiting.
        sys.exit(returncode)


def parse_args(argv: list[str] | None = None) -> argparse.Namespace:
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(
        prog="rr-compile-tectonic",
        description="Compile a LaTeX document with Tectonic and deduce input and output info.",
    )
    parser.add_argument("path_tex", type=Path, help="The main LaTeX source file.")
    parser.add_argument(
        "-m",
        "--maxrep",
        default=6,
        type=int,
        help="The maximum number of LaTeX recompilations (not including the one for BibTeX).",
    )
    parser.add_argument(
        "--tectonic",
        help="The Tectonic executable. "
        "The default is ${REPREP_TECTONIC} or tectonic if the variable is not defined.",
    )
    parser.add_argument(
        "--keep-deps",
        help="Keep the dependency file after the compilation. "
        "The default is to use a temporary file, which is removed after it is processed.",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--inventory",
        type=Path,
        help="Write an inventory with all inputs and outputs, useful for archiving.",
    )
    parser.add_argument(
        "tectonic_args",
        nargs="*",
        help="Additional arguments to be passed to tectonic. "
        "The defaults is `${REPREP_TECTONIC_ARGS}`, if the environment variable is defined.",
    )
    return parser.parse_args(argv)


if __name__ == "__main__":
    main(sys.argv[1:])



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# © 2024–2025 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""RepRep Wrapper for LaTeX."""

import argparse
import contextlib
import shlex
import sys

from path import Path

from stepup.core.api import amend, getenv
from stepup.core.hash import compute_file_digest
from stepup.core.utils import filter_dependencies, mynormpath
from stepup.core.worker import WorkThread

from .bibtex_log import parse_bibtex_log
from .latex_deps import scan_latex_deps
from .latex_log import parse_latex_log
from .make_inventory import write_inventory


def main(argv: list[str] | None = None, work_thread: WorkThread | None = None) -> None:
    """Main program."""
    args = parse_args(argv)
    if work_thread is None:
        work_thread = WorkThread("stub")

    workdir, fn_tex = args.path_tex.splitpath()
    workdir = workdir.normpath()
    if not fn_tex.endswith(".tex"):
        raise ValueError("The LaTeX source must have extension .tex")
    stem = fn_tex[:-4]
    path_aux = workdir / f"{stem}.aux"

    # Remove existing outputs from a previous run, which could potentially
    # conflict with the new tex source files. In 99% of the cases, this is
    # not a problem, but sometimes LaTeX chokes on remnants in old outputs.
    exts_to_remove = ["log", "aux", "blg", "fls", "out", "toc", "nlo", "synctex"]
    if args.run_bibtex:
        exts_to_remove.append("bbl")
    for ext in exts_to_remove:
        (workdir / f"{stem}.{ext}").remove_p()

    # Detect additional inputs
    inp, bib, out, vol = scan_latex_deps(fn_tex, do_amend=False)

    # Get LaTeX executable
    if args.latex is None:
        args.latex = getenv("REPREP_LATEX", "pdflatex")

    aux_digest_hist = []
    if len(bib) == 0:
        amend(inp=inp, out=out, vol=vol)
        inventory_files = [*inp, *out]
    elif args.run_bibtex:
        # Get other executables and files
        if args.bibtex is None:
            args.bibtex = getenv("REPREP_BIBTEX", "bibtex")

        amend(inp=inp + bib, out=[f"{stem}.bbl", *out], vol=vol)
        inventory_files = [*inp, *bib, f"{stem}.bbl", *out]

        # Run LaTeX once to generate the .aux file
        with contextlib.chdir(workdir):
            returncode, _, _ = work_thread.runsh(
                f"{shlex.quote(args.latex)} -recorder -interaction=errorstopmode -draftmode {stem}"
            )
        if returncode != 0:
            path_log = workdir / f"{stem}.log"
            error_info = parse_latex_log(path_log)
            error_info.print(path_log)
            sys.exit(1)

        aux_digest_hist.append(compute_file_digest(path_aux))

        # BibTeX
        with contextlib.chdir(workdir):
            returncode, _, _ = work_thread.runsh(f"{shlex.quote(args.bibtex)} {stem}")
        if returncode != 0:
            path_blg = workdir / f"{stem}.blg"
            error_info = parse_bibtex_log(path_blg)
            error_info.print(path_blg)
            sys.exit(1)
    else:
        amend(inp=[*inp, f"{stem}.bbl"], out=out, vol=vol)
        inventory_files = [*inp, f"{stem}.bbl", *out]

    # Keep running LaTeX until the .aux file converges.
    for _ in range(args.maxrep):
        with contextlib.chdir(workdir):
            returncode, _, _ = work_thread.runsh(
                f"{shlex.quote(args.latex)} -recorder -interaction=errorstopmode {stem}"
            )
        path_log = workdir / f"{stem}.log"
        error_info = parse_latex_log(path_log)
        if returncode != 0:
            error_info.print(path_log)
            sys.exit(1)
        aux_digest_hist.append(compute_file_digest(path_aux))
        if len(aux_digest_hist) > 1 and aux_digest_hist[-1] == aux_digest_hist[-2]:
            break
    else:
        print(
            f"\033[1;31;40mAux file did not converge in {args.maxrep} iterations!\033[0;0m",
            file=sys.stderr,
        )
        print(path_aux, file=sys.stderr)
        for digest in aux_digest_hist:
            print(digest.hex(), file=sys.stderr)
        sys.exit(1)

    # Write inventory
    inventory_files.extend([f"{stem}.tex", f"{stem}.aux", f"{stem}.pdf"])
    if args.inventory is not None:
        write_inventory(args.inventory, inventory_files, do_amend=False)

    # Look for input files and output files from the fls file.
    # These are usually worth tracking, but are not needed for the inventory file.
    fls_inp = set()
    fls_vol = set()
    with open(f"{stem}.fls") as fh:
        for line in fh:
            if line.startswith("INPUT "):
                path = mynormpath(Path(line[6:].strip()))
                if not (path in inventory_files or path == args.inventory):
                    fls_inp.add(path)
            elif line.startswith("OUTPUT "):
                path = mynormpath(Path(line[7:].strip()))
                if not (path in inventory_files or path == args.inventory):
                    fls_vol.add(path)
    fls_inp.difference_update(fls_vol)
    # Both inputs and outputs must be filtered because, strangely,
    # LaTeX sometimes outputs files in the weirdest places, e.g. in the TEXMF tree.
    amend(inp=filter_dependencies(fls_inp), vol=filter_dependencies(fls_vol))


def parse_args(argv: list[str] | None = None) -> argparse.Namespace:
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(
        prog="rr-compile-latex",
        description="Compile a LaTeX document and extract input and output info.",
    )
    parser.add_argument("path_tex", type=Path, help="The main LaTeX source file.")
    parser.add_argument(
        "-m",
        "--maxrep",
        default=5,
        type=int,
        help="The maximum number of LaTeX recompilations (not including the one for BibTeX).",
    )
    parser.add_argument(
        "--latex",
        help="The LaTeX executable. "
        "The default is ${REPREP_LATEX} or pdflatex if the variable is not defined.",
    )
    parser.add_argument(
        "--run-bibtex",
        dest="run_bibtex",
        default=False,
        action="store_true",
        help="Run bibtex.",
    )
    parser.add_argument(
        "--bibtex",
        help="The BibTeX executable. "
        "The default is ${REPREP_BIBTEX} or bibtex if the variable is not defined.",
    )
    parser.add_argument(
        "--inventory",
        type=Path,
        help="Write an inventory with all inputs and outputs, useful for archiving.",
    )
    return parser.parse_args(argv)


if __name__ == "__main__":
    main(sys.argv[1:])



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# © 2024–2025 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Extract data points from plots."""

import argparse
import json
from xml.etree.ElementTree import Element

import attrs
import numpy as np
from defusedxml.ElementTree import iterparse
from numpy.typing import NDArray
from path import Path
from svg.path import parse_path


def main(argv: list[str] | None = None):
    """Main program."""
    args = parse_args(argv)
    axes, px_curves = load_pixel_data_svg(args.inp)
    data = {
        "units": {axis.label: axis.units for axis in axes},
        "curves": {
            label: transform_pixel_data(axes, px_curve) for label, px_curve in px_curves.items()
        },
    }
    with open(args.out, "w") as fh:
        json.dump(data, fh, indent=2)
        fh.write("\n")


def parse_args(argv: list[str] | None = None) -> argparse.Namespace:
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(
        prog="rr-unplot", description="Extract data points from bitmap plots."
    )
    parser.add_argument("inp", type=Path, help="SVG file with paths drawn on top of the image")
    parser.add_argument("out", type=Path, help="Output JSON file with extracted data points")
    return parser.parse_args(argv)


@attrs.define
class Axis:
    """Extracted data of an x- or y-axis."""

    start: NDArray = attrs.field()
    end: NDArray = attrs.field()
    label: str = attrs.field()
    low: float = attrs.field(converter=float)
    high: float = attrs.field(converter=float)
    scale: str = attrs.field(validator=attrs.validators.in_(["linear", "log"]))
    units: str = attrs.field()


def load_pixel_data_svg(filename: str) -> tuple[list[Axis], dict[str, NDArray]]:
    """Load elements from an SVG file, keeping all data in the units of the SVG drawing.

    Parameters
    ----------
    filename
        The SVG file to load.

    Returns
    -------
    axes
        The two axes of the plot.
    pixel_curves
        A dictionary with curve data loaded from the plot.
        The keys are labels extracted from the path id.
        The values are NumPy arrays of which the rows are (x, y) data points.
    """
    px_curves = {}
    axes = []
    parser = iterparse(filename, events=("start",))
    for event, elem in parser:
        if event == "start":
            tag = elem.tag.rpartition("}")[2]
            if tag == "path":
                if "axis" in elem.attrib:
                    axes.append(parse_axis(elem))
                elif "data" in elem.attrib:
                    label = elem.attrib["data"]
                    px_curve = parse_curve(elem)
                    px_curves[label] = px_curve
        elem.clear()

    if len(axes) != 2:
        raise ValueError(f"Expecting two axes, got {len(axes)}")
    if len(px_curves) == 0:
        raise ValueError("No curve with data points found.")
    return axes, px_curves


def parse_axis(path: Element) -> Axis:
    """Convert an SVG Path to an x- or y-axis."""
    id_ = path.attrib["id"]
    nodes = extract_svg_path(path.attrib["d"])
    if len(nodes) != 2:
        raise ValueError(f"Expected two nodes, got {len(nodes)} ({id_})")
    label = path.attrib.get("axis")
    if label is None:
        raise ValueError(f"An axis has no required axis attribute ({id_})")
    low = path.attrib["low"]
    high = path.attrib["high"]
    scale = path.attrib["scale"]
    units = path.attrib["units"]
    return Axis(nodes[0], nodes[1], label, low, high, scale, units)


def parse_curve(path: Element) -> NDArray:
    """Convert an SVG Path to an array of data points."""
    return extract_svg_path(path.attrib["d"])


def extract_svg_path(d: str) -> NDArray:
    """Convert SVG path commands into an array of data points."""
    path = parse_path(d)
    result = []
    for item in path:
        for point in item.start, item.end:
            if len(result) == 0 or result[-1] != point:
                result.append(point)
    return np.array([[point.real, point.imag] for point in result])


def transform_pixel_data(axes: list[Axis], px_curve: NDArray) -> NDArray:
    """Convert data in drawing coordinates to axes coordinates."""
    if len(axes) != 2:
        raise ValueError(f"Expecting two axes, got {len(axes)}")
    # The labels x and y are a bit arbitrary,
    # because any affine transformation of the axes is supported.
    x_axis, y_axis = axes

    # Construct x- and y-unit vectors in pixel coordinates.
    px_xunit = x_axis.end - x_axis.start
    px_yunit = y_axis.end - y_axis.start

    # The affine transformation to pixel coordinates.
    mat_to_pix = np.array([px_xunit, px_yunit]).T

    # The inverse affine transformation.
    mat_from_pix = np.linalg.inv(mat_to_pix)

    # Reference point for the x and y values.
    x_low = np.dot(mat_from_pix, x_axis.start)
    y_low = np.dot(mat_from_pix, y_axis.start)

    # Transform the datapoints to data coordinates.
    data = np.dot(mat_from_pix, px_curve.T)
    data[0] -= x_low[0]
    data[1] -= y_low[1]

    # Convert to plot units.
    convert_unit(data[0], x_axis)
    convert_unit(data[1], y_axis)

    # Associate the data with the axis labels, instead of x and y.
    return {x_axis.label: data[0].tolist(), y_axis.label: data[1].tolist()}


def convert_unit(values: NDArray, axis: Axis):
    """Transform data in place to axis coordinates.

    The values are assumed to be transformed to dimensionless coordintes first.
    """
    if axis.scale == "linear":
        values[:] = values * (axis.high - axis.low) + axis.low
    elif axis.scale == "log":
        llow = np.log(axis.low)
        lhigh = np.log(axis.high)
        values[:] = np.exp(values * (lhigh - llow) + llow)
    else:
        raise ValueError(f"Unsupported axis scale: {axis.scale}")


if __name__ == "__main__":
    main()



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# © 2024–2025 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Offline checker for hyper references."""

import argparse
import re
import sys

import attrs
import cattrs.preconf.pyyaml
import fitz
import yaml
from bs4 import BeautifulSoup
from markdown_it import MarkdownIt
from path import Path

from stepup.core.api import amend, getenv
from stepup.core.utils import mynormpath


def main(argv: list[str] | None = None):
    """Main program."""
    args = parse_args(argv)
    if args.path_config is None:
        args.path_config = getenv("REPREP_CHECK_HREFS_CONFIG", "check_hrefs.yaml", back=True)
        amend(inp=args.path_config)
    config = load_config(args.path_config)
    hrefs = collect_hrefs(args.path_src)
    make_url_substitutions(hrefs, config.subs)
    check_hrefs(hrefs, Path(args.path_src).parent.normpath(), config.accept)


def parse_args(argv: list[str] | None = None) -> argparse.Namespace:
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(
        prog="rr-check-hrefs", description="Check hyper references Markdown, HTML or PDF files."
    )
    parser.add_argument("path_src", help="Markdown, HTML or PDF file.")
    parser.add_argument(
        "-c",
        "--config",
        dest="path_config",
        help="Configuration yaml file. "
        "The default is ${REPREP_CHECK_HREFS_CONFIG} or check_hrefs.yaml if it is not set.",
    )
    return parser.parse_args(argv)


@attrs.define
class Config:
    accept: set[str] = attrs.field(factory=set)
    subs: list[tuple[str, str]] = attrs.field(factory=list)


def load_config(path_config) -> Config:
    converter = cattrs.preconf.pyyaml.PyyamlConverter(forbid_extra_keys=True)
    with open(path_config) as fh:
        state = yaml.safe_load(fh)
        return converter.structure(state, Config)


@attrs.define
class HRef:
    """A hyper reference to be checked."""

    url: str = attrs.field()
    translated: str = attrs.field(default=None)
    ignore: bool = attrs.field(default=False, init=False)


def collect_hrefs(path_src: str) -> list[HRef]:
    """Find all hyper references in one file."""
    if path_src.endswith(".md"):
        result = collect_hrefs_md(path_src)
    elif path_src.endswith(".html"):
        result = collect_hrefs_html(path_src)
    elif path_src.endswith(".pdf"):
        result = collect_hrefs_pdf(path_src)
    else:
        raise ValueError(f"Source file type not supported: {path_src}")
    return result


def collect_hrefs_md(fn_md: str) -> list[HRef]:
    """Find all hyper references in a Markdown file."""
    md = MarkdownIt()
    with open(fn_md) as fh:
        html = md.render(fh.read())
    soup = BeautifulSoup(html, "html.parser")
    return [HRef(link.attrs["href"]) for link in soup.find_all("a")]


def collect_hrefs_html(fn_html: str) -> list[HRef]:
    """Find all hyper references in a HTML file."""
    with open(fn_html) as fh:
        soup = BeautifulSoup(fh.read(), "html.parser")
    return [HRef(link.attrs["href"]) for link in soup.find_all("a")]


def collect_hrefs_pdf(fn_pdf: str) -> list[HRef]:
    """Find all hyper references in a PDF file."""
    # See https://pymupdf.readthedocs.io/en/latest/page.html#description-of-get-links-entries
    doc = fitz.open(fn_pdf)
    hrefs = []
    for page in doc:
        for link in page.get_links():
            if link["kind"] in (fitz.LINK_LAUNCH, fitz.LINK_GOTOR):
                hrefs.append(HRef(link["file"]))
            elif link["kind"] == fitz.LINK_URI:
                hrefs.append(HRef(link["uri"]))
    return hrefs


def make_url_substitutions(hrefs: list[HRef], subs: list[tuple[str, str]]):
    for href in hrefs:
        url = href.url
        if "#" in href.url:
            url = url[: href.url.find("#")]
        for pattern, replacement in subs:
            url = re.sub(pattern, replacement, url)
        href.translated = url


def check_hrefs(hrefs: list[HRef], root: Path, accept: set[str]):
    """Check the hyper references."""
    seen = set()
    some_failed = False
    for href in hrefs:
        if href.url in seen:
            continue
        some_failed |= check_href(href, root, accept)
        seen.add(href.url)
    if some_failed:
        raise ValueError("Some hyper references were invalid.")


def check_href(href: HRef, root: Path, accept: set[str]) -> bool:
    """Check a hyper reference and return True if an error was found."""
    if href.url.startswith("mailto:"):
        return False
    if "://" in href.translated:
        if href.translated in accept:
            return False
        print(f"FAILED LINK: {href.url}", file=sys.stderr)
        return True
    path = mynormpath(root / Path(href.translated))
    amend(inp=path)
    return False


if __name__ == "__main__":
    sys.exit(main())



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# © 2024–2025 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Concatenate multiple PDFs into a single document, optionally inserting blank pages."""

import argparse
import sys

import fitz

__all__ = ("cat_pdf",)


def main(argv: list[str] | None = None):
    """Main program."""
    args = parse_args(argv)
    cat_pdf(args.paths_src, args.path_dst, args.insert_blank)


def parse_args(argv: list[str] | None = None) -> argparse.Namespace:
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(
        prog="rr-cat-pdf", description="Concatenate PDFs, optionally inserting blank pages."
    )
    parser.add_argument(
        "paths_src", help="The source PDFs to which notes should be added.", nargs="+"
    )
    parser.add_argument("path_dst", help="The output PDF.")
    parser.add_argument(
        "-i",
        "--insert-blank",
        help="Insert a blank page after a PDF with an odd number of pages. "
        "The last page of each PDF is used to determine the size of the added blank page.",
        default=False,
        action="store_true",
    )
    return parser.parse_args(argv)


def cat_pdf(
    paths_src: list[str],
    path_dst: str,
    insert_blank: bool,
):
    """Put multiple pages in a single page, using a fixed layout.

    Parameters
    ----------
    paths_src
        The source PDF filename.
    path_dst
        The destination PDF filename.
    insert_blank
        Insert a blank page after a PDF with an odd number of pages.
        The last page of each PDF is used to determine the size of the added blank page.
    """
    for path_pdf in [*paths_src, path_dst]:
        if not path_pdf.endswith(".pdf"):
            raise ValueError(
                f"All arguments must have a `.pdf` extension, got: {path_pdf}", file=sys.stderr
            )
    dst = fitz.open()

    for path_src in paths_src:
        src = fitz.open(path_src)
        # See https://github.com/pymupdf/PyMuPDF/issues/3635
        src.scrub()
        dst.insert_pdf(src)
        if insert_blank and src.page_count % 2 == 1:
            last_page = src[-1]
            dst.insert_page(-1, width=last_page.rect.width, height=last_page.rect.height)
        src.close()

    # Strip metadata for reproducibility and save
    dst.set_metadata({})
    dst.del_xml_metadata()
    dst.xref_set_key(-1, "ID", "null")
    dst.scrub()
    dst.save(path_dst, garbage=4, deflate=True, no_new_id=True)

    dst.close()


if __name__ == "__main__":
    main(sys.argv[1:])



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# © 2024–2025 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Wrapper for HTML to PDF conversion."""

import argparse
import shlex
import sys
from collections.abc import Iterator

from defusedxml import ElementTree
from path import Path

from stepup.core.api import amend, getenv
from stepup.core.worker import WorkThread


def main(argv: list[str] | None = None, work_thread: WorkThread | None = None):
    """Main program."""
    args = parse_args(argv)
    if work_thread is None:
        work_thread = WorkThread("stub")

    if not args.path_pdf.endswith(".pdf"):
        raise ValueError("The output must have a pdf extensions.")
    if args.weasyprint is None:
        args.weasyprint = getenv("REPREP_WEASYPRINT", "weasyprint")
    inp_paths = search_html_deps(args.path_html)
    amend(inp=inp_paths)
    popenargs = [args.weasyprint, args.path_html, args.path_pdf]
    returncode = work_thread.runsh_verbose(shlex.join(popenargs))
    if returncode != 0:
        sys.exit(returncode)


def parse_args(argv: list[str] | None = None) -> argparse.Namespace:
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(
        prog="rr-convert-weasyprint",
        description="Convert a HTML to PDF, with dependency tracking.",
    )
    parser.add_argument("path_html", help="The input HTML file.")
    parser.add_argument("path_pdf", help="The output PDF file.")
    parser.add_argument(
        "--weasyprint",
        help="The weasyprint executable to use. "
        "Defaults to `${REPREP_WEASYPRINT}` variable or `weasyprint` if the variable is unset.",
    )
    return parser.parse_args(argv)


def search_html_deps(src: str) -> list[str]:
    """Search implicit dependencies in HTML files: included images and css."""
    implicit = []
    todo = [src]
    idep = 0
    while idep < len(todo):
        path_html = Path(todo[idep])
        for href in iter_html_hrefs(path_html):
            if href.startswith("file://"):
                href = href[7:]
            if "://" not in href:
                if not href.startswith("/"):
                    href = path_html.parent / href
                implicit.append(href)
                if href.endswith(".html"):
                    todo.append(href)
        idep += 1
    return implicit


def iter_html_hrefs(path_html: str) -> Iterator[str]:
    parser = ElementTree.iterparse(path_html, events=("start",))
    for event, elem in parser:
        if event == "start":
            tag = elem.tag.rpartition("}")[2]
            if tag == "img":
                for key in "src", "{http://www.w3.org/1999/xhtml}src":
                    href = elem.attrib.get(key)
                    if href is not None and "data:image" not in href:
                        yield href
                        break
            elif tag == "link":
                for key in "href", "{http://www.w3.org/1999/xhtml}href":
                    href = elem.attrib.get(key)
                    if href is not None and "data:image" not in href:
                        yield href
                        break
        elem.clear()


if __name__ == "__main__":
    main(sys.argv[1:])



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# © 2024–2025 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Tile PDF figures."""

from collections.abc import Collection

import attrs
import fitz
import numpy as np
from scipy import optimize

__all__ = ("Figure", "Panel")


@attrs.define
class Panel:
    """The definition of one panel in a tiled figure.

    Attributes
    ----------
    irow
        The row where the panel is located (top left corner).
    icol
        The column where the panel is located (top left corner).
    label
        The label to be put above the panel.
    path_in
        The path of the PDF file with the figure.
    nrow
        The number of rows occupied by the panel.
    ncol
        The number of columns occupied by the panel.
    """

    irow: int = attrs.field()
    icol: int = attrs.field()
    label: str = attrs.field()
    path_in: str = attrs.field()
    nrow: int = attrs.field(default=1)
    ncol: int = attrs.field(default=1)
    pdf = attrs.field(default=None)


@attrs.define
class Figure:
    """The definition of a tiled figure

    Attributes
    ----------
    path_out
        The PDF output file.
    panels
        The list panels, instances of the `Panel` class.
    fontname
        A Fontname recognized by PyMyPDF or a custom name when fontfile is specified.
    fontfile
        None or the path to a ttf file.
        When used, specify a corresponding fontname (of your choice).
    fontsize
        The font size to use for the labels in points.
    label_height
        The height to use for the labels in mm.
    padding
        The padding added added to the panels before combining them, in mm.
        This parameter is also used as margin between the label and the figure.
    hshift
        An optional horizontal displacement of the panel label, in mm.
    """

    path_out: str = attrs.field()
    panels: list[Panel] = attrs.field()
    fontname: str = attrs.field(default="hebo")
    fontfile: str | None = attrs.field(default=None)
    # Dimensions in points. (1 point = 1/72 inch)
    fontsize: float = attrs.field(default=7.0)
    label_height: float = attrs.field(default=2.0)
    padding: float = attrs.field(default=1.0)
    hshift: float = attrs.field(default=0.0)

    def info(self):
        inp = [panel.path_in for panel in self.panels]
        if self.fontfile is not None:
            inp.append(self.fontfile)
        return {"inp": inp, "out": self.path_out}

    def run(self):
        """Combine PDF figures into a single PDF with labels on top of each panel."""
        _load_pdfs(self.panels)
        for panel in self.panels:
            _add_label(panel, self)
        out = _combine_figures(self.panels)
        out.set_metadata({})
        out.del_xml_metadata()
        out.scrub()
        out.save(self.path_out, garbage=4, deflate=True, no_new_id=True)


def _load_pdfs(panels: Collection[Panel]):
    for panel in panels:
        panel.pdf = fitz.open(panel.path_in)
        # See https://github.com/pymupdf/PyMuPDF/issues/3635
        panel.pdf.scrub()
        if panel.pdf.page_count != 1:
            raise ValueError(
                "Panel PDF files should have just one page. "
                f"Found {panel.pdf.page_count} in {panel.path_in}"
            )


mm = 72 / 25.4


def _add_label(panel: Panel, fig: Figure):
    new = fitz.open()
    old_page = panel.pdf[0]
    new_page = new.new_page(
        width=old_page.rect.width + 2 * fig.padding * mm,
        height=old_page.rect.height + fig.label_height * mm + 3 * fig.padding * mm,
    )
    top = fig.label_height * mm + 2 * fig.padding * mm
    new_page.show_pdf_page(
        fitz.Rect(
            fig.padding * mm,
            top,
            old_page.rect.width + fig.padding * mm,
            old_page.rect.height + top,
        ),
        panel.pdf,
        0,
    )
    font = fitz.Font(fig.fontname, fig.fontfile)
    length = font.text_length(panel.label, fontsize=fig.fontsize)
    text_writer = fitz.TextWriter(new_page.rect)
    text_writer.append(
        fitz.Point(
            (new_page.rect.width - length) / 2 + fig.hshift * mm,
            fig.padding * mm + fig.label_height * mm,
        ),
        panel.label,
        font=font,
        fontsize=fig.fontsize,
    )
    text_writer.write_text(new_page)
    panel.pdf = new


def _combine_figures(panels: list[Panel]):
    # Basic settings
    nrow = 1
    ncol = 1
    for sf in panels:
        nrow = max(nrow, sf.irow + sf.nrow)
        ncol = max(ncol, sf.icol + sf.ncol)

    # Define optimization variables
    row_vars = np.arange(nrow)
    col_vars = np.arange(ncol) + nrow
    nvar = nrow + ncol

    # Run over panels and define constraints
    a_ub = np.zeros((2 * len(panels), nvar))
    b_ub = np.zeros(2 * len(panels))
    c = np.zeros(nvar)
    c[row_vars[-1]] = 1
    c[col_vars[-1]] = 1
    ieq = 0
    for sf in panels:
        if sf.irow == 0:
            a_ub[ieq, row_vars[sf.nrow - 1]] = -1
        else:
            a_ub[ieq, row_vars[sf.irow + sf.nrow - 1]] = -1
            a_ub[ieq, row_vars[sf.irow - 1]] = 1
        b_ub[ieq] = -sf.pdf[0].rect.height
        ieq += 1
        if sf.icol == 0:
            a_ub[ieq, col_vars[sf.ncol - 1]] = -1
        else:
            a_ub[ieq, col_vars[sf.icol + sf.ncol - 1]] = -1
            a_ub[ieq, col_vars[sf.icol - 1]] = 1
        b_ub[ieq] = -sf.pdf[0].rect.width
        ieq += 1

    # Optimize the layout
    res = optimize.linprog(c, a_ub, b_ub, method="highs")
    row_pos = np.concatenate([[0.0], res.x[:nrow]])
    col_pos = np.concatenate([[0.0], res.x[nrow:]])

    # Put everything in one PDF
    out = fitz.open()
    page = out.new_page(width=col_pos[-1], height=row_pos[-1])
    for sf in panels:
        dst_rect = fitz.Rect(
            col_pos[sf.icol],
            row_pos[sf.irow],
            col_pos[sf.icol + sf.ncol],
            row_pos[sf.irow + sf.nrow],
        )
        page.show_pdf_page(dst_rect, sf.pdf, 0)
    return out



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# © 2024–2025 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Create ZIP with all files listed in an inventory.txt file."""

import argparse
import datetime
import sys
import tempfile
import zipfile

from path import Path

from .check_inventory import iter_inventory
from .inventory import check_summary, get_summary

__all__ = ("zip_inventory",)


TIMESTAMP = datetime.datetime(1980, 1, 1).timestamp()


def main(argv: list[str] | None = None):
    """Main program."""
    parser = argparse.ArgumentParser(
        prog="rr-zip-inventory", description="Create a reproducible ZIP file."
    )
    add_parser_args(parser)
    args = parser.parse_args(argv)
    zip_inventory(args.inventory_txt, args.output_zip)


def add_parser_args(parser: argparse.ArgumentParser):
    """Define command-line arguments."""
    parser.add_argument(
        "inventory_txt",
        help="The inventory file with all files to be zipped. "
        "The digests of the files will be checked before archiving. "
        "The inventory file will be included in the ZIP.",
    )
    parser.add_argument("output_zip", nargs="?", help="Destination zip file.")


def zip_subcommand(subparser: argparse.ArgumentParser) -> callable:
    parser = subparser.add_parser(
        "zip-inventory",
        help="Create a reproducible ZIP file.",
    )
    add_parser_args(parser)
    return zip_tool


def zip_tool(args: argparse.Namespace) -> int:
    """Create a reproducible ZIP file."""
    zip_inventory(args.inventory_txt, args.output_zip)
    return 0


def zip_inventory(path_inventory: str, path_zip: str | None = None):
    """Create a reproducible zip file.

    Parameters
    ----------
    path_inventory
        The inventory.txt file.
        All files listed in the inventory are checked and then zipped.
    path_zip
        The ZIP file to be created.
        When not given, the `.zip` suffix is added to the prefix of `path_inventory`.
        The existing ZIP file with the same path is only overwritten when the ZIP
        file is first succesfully created in a temporary directory.
    """
    if not path_inventory.endswith(".txt"):
        raise ValueError(f"The inventory file must have a `.txt` extension. Got {path_inventory}")
    path_inventory = Path(path_inventory)
    if path_zip is None:
        path_zip = path_inventory[:-4] + ".zip"
    elif not path_zip.endswith(".zip"):
        raise ValueError(f"Destination must have a `.zip` extension. Got {path_zip}")
    path_zip = Path(path_zip)

    # Create a new ZIP archive.
    root = path_inventory.parent
    nskip = 0 if root == "" else len(root) + 1
    with tempfile.TemporaryDirectory("rr-zip-inventory") as path_tmp:
        path_tmp = Path(path_tmp)
        path_zip_tmp = path_tmp / "out.zip"
        with zipfile.ZipFile(path_zip_tmp, "w") as fz:
            for ref in iter_inventory(path_inventory):
                # Copy the file to a temp dir.
                # This saves bandwith in case of remote datasets and allows
                # fixing the timestamp before compression.
                src = Path(root / ref.path)
                dst = path_tmp / "todo"
                dst.remove_p()
                src.copy(dst, follow_symlinks=False)
                # Check the file before adding it to the ZIP
                new = get_summary(dst, path_tmp)
                check_summary(new, ref)
                # Store
                if ref.size is None:
                    # Symbolic link
                    zipinfo = zipfile.ZipInfo(src[nskip:])
                    zipinfo.create_system = 3  # 3 means Unix
                    zipinfo.external_attr |= dst.stat(follow_symlinks=False).st_mode << 16
                    fz.writestr(zipinfo, dst.readlink())
                else:
                    # Actual file
                    dst.utime((TIMESTAMP, TIMESTAMP))
                    fz.write(dst, src[nskip:], zipfile.ZIP_DEFLATED)
            dst = path_tmp / "todo"
            path_inventory.copy(dst)
            dst.utime((TIMESTAMP, TIMESTAMP))
            fz.write(dst, path_inventory[nskip:], zipfile.ZIP_DEFLATED)
        path_zip_tmp.move(path_zip)


if __name__ == "__main__":
    main(sys.argv[1:])



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# © 2024–2025 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Wrapper for SVG to PDF conversion."""

import argparse
import os
import shlex
import sys
from collections.abc import Iterator

from defusedxml import ElementTree
from path import Path

from stepup.core.api import amend, getenv
from stepup.core.utils import filter_dependencies
from stepup.core.worker import WorkThread


def main(argv: list[str] | None = None, work_thread: WorkThread | None = None):
    """Main program."""
    args = parse_args(argv)
    if work_thread is None:
        work_thread = WorkThread("stub")

    path_out = Path(args.path_out)
    allowed_extensions = [".pdf", ".png"]
    if not any(path_out.endswith(ext) for ext in allowed_extensions):
        raise ValueError(
            f"The output must have one of the following extensions: {allowed_extensions}"
        )
    if args.inkscape is None:
        args.inkscape = getenv("REPREP_INKSCAPE", "inkscape")
    if len(args.inkscape_args) == 0:
        args.inkscape_args = shlex.split(
            getenv(f"REPREP_INKSCAPE_{path_out.suffix[1:].upper()}_ARGS", "")
        )
    inp_paths = filter_dependencies(search_svg_deps(args.path_svg))
    amend(inp=inp_paths)
    args = [
        args.inkscape,
        args.path_svg,
        *args.inkscape_args,
        f"--export-filename={path_out}",
        f"--export-type={path_out.suffix[1:]}",
    ]
    os.environ["SELF_CALL"] = "x"
    returncode = work_thread.runsh_verbose(shlex.join(args))
    if returncode != 0:
        sys.exit(returncode)


def parse_args(argv: list[str] | None = None) -> argparse.Namespace:
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(
        prog="rr-convert-inkscape",
        description="Convert an SVG to PDF or PNG, with dependency tracking.",
    )
    parser.add_argument("path_svg", help="The input SVG file.")
    parser.add_argument("path_out", help="The output PDF or PNG file.")
    parser.add_argument(
        "inkscape_args",
        nargs="*",
        help="Additional arguments to be passed to inkscape. "
        "E.g. -T (for text to path conversion). "
        "Depending on the output extension, the defaults is `${REPREP_INKSCAPE_PDF_ARGS}` or "
        "`${REPREP_INKSCAPE_PDF_ARGS}`, if the environment variable is defined.",
    )
    parser.add_argument(
        "--inkscape",
        help="The inkscape executable to use. "
        "Defaults to `${REPREP_INKSCAPE}` variable or `inkscape` if the variable is unset.",
    )
    return parser.parse_args(argv)


def search_svg_deps(src: str) -> list[str]:
    """Search implicit dependencies in SVG files, specifically (recursively) included images."""
    implicit = []
    todo = [src]
    idep = 0
    while idep < len(todo):
        path_svg = Path(todo[idep])
        for href in iter_svg_image_hrefs(path_svg):
            if href.startswith("file://"):
                href = href[7:]
            if "://" not in href:
                if not href.startswith("/"):
                    href = path_svg.parent / href
                implicit.append(href)
                if href.endswith(".svg"):
                    todo.append(href)
        idep += 1
    return implicit


def iter_svg_image_hrefs(path_svg: str) -> Iterator[str]:
    parser = ElementTree.iterparse(path_svg, events=("start",))
    for event, elem in parser:
        if event == "start":
            tag = elem.tag.rpartition("}")[2]
            if tag == "image":
                for key in "href", "{http://www.w3.org/1999/xlink}href":
                    href = elem.attrib.get(key)
                    if href is not None and "data:image" not in href:
                        yield href
                        break
        elem.clear()


if __name__ == "__main__":
    main(sys.argv[1:])



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# © 2024–2025 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
r"""Recursively flatten a LaTeX file.

The following commands are expanded:

- ``\input``: standard LaTeX command
- ``\import``: like ``\input``, but capable of handling directory changes.

The following commands are rewritten to be consistent with the location of the flattened output:

- ``\includegraphics``
- ``\thebibliography``
- ``\verbatiminput``

This script is intentionally somewhat limited.
It expects that ``\input`` and ``\import`` commands are the only ones present on their line,
to avoid ambiguities. If this is not the case, the script will fail.
The script also assumes the ``\includgraphics``, ``\thebibliography`` and ``\verbatiminput``
commands are contained within a single line.
"""

import argparse
import enum
import re
import sys
import tempfile
from typing import TextIO

from path import Path

from stepup.core.api import amend


def main(argv: list[str] | None = None):
    """Main program."""
    args = parse_args(argv)
    with tempfile.TemporaryDirectory("rr-latex-flat") as tmpdir:
        tmpdir = Path(tmpdir)
        path_flat_tmp = tmpdir / "flat.tex"
        with open(path_flat_tmp, "w") as fh_out:
            out_root = Path(args.path_flat).parent
            status, inp_paths = flatten_latex(args.path_tex, fh_out, out_root)
        amend(inp=inp_paths)
        if status == FlattenStatus.SUCCESS:
            path_flat_tmp.copy(args.path_flat)
        else:
            raise RuntimeError(f"Flattening failed with status {status.name}. ")


def parse_args(argv: list[str] | None = None) -> argparse.Namespace:
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(
        prog="rr-flatten-latex",
        description="Flatten input and import commands in a LaTeX file.",
    )
    parser.add_argument("path_tex", help="The top-level tex file.")
    parser.add_argument("path_flat", help="The flattened output tex file.")
    return parser.parse_args(argv)


class FlattenStatus(enum.Enum):
    SUCCESS = 0
    FILE_NOT_FOUND = 1
    ILL_FORMATTED = 2


def flatten_latex(
    path_tex: str, fh_out: TextIO, out_root: str, tex_root: str | None = None
) -> tuple[FlattenStatus, list[str]]:
    """Write a flattened LaTeX file

    Parameters
    ----------
    path_tex
        The LaTeX source to be flattened, may be the main file or
        an included file in one of the recursions.
    fh_out
        The file object to write the flattened file to.
    out_root
        The directory of the output file, needed to fix relative paths.
    tex_root
        The directory with respect to which paths in the LaTeX source must
        be interpreted.

    Returns
    -------
    success
        True when successful.
    inp_paths
        A list of additional inputs used.
    """
    inp_paths = []
    path_tex = Path(path_tex)
    tex_root = path_tex.parent.normpath() if tex_root is None else Path(tex_root)
    status = FlattenStatus.SUCCESS
    with open(path_tex) as fh:
        for iline, line in enumerate(fh):
            # Reduce line to standard form
            stripped = line[: line.find("%")].strip()
            stripped = stripped.replace(" ", "").replace("\t", "")

            # Try to find input or import
            sub_path_tex = None
            new_root = tex_root
            if r"\input{" in stripped:
                if stripped.startswith(r"\input{") and stripped.endswith("}"):
                    sub_path_tex = tex_root / stripped[7:-1]
                    if sub_path_tex.suffix == "":
                        sub_path_tex = sub_path_tex.with_suffix(".tex")
                    if not sub_path_tex.is_file():
                        status = FlattenStatus.FILE_NOT_FOUND
                else:
                    status = FlattenStatus.ILL_FORMATTED
            elif r"\import{" in stripped:
                if stripped.startswith(r"\import{") and "}{" in stripped and stripped.endswith("}"):
                    new_root, sub_path_tex = stripped[8:-1].split("}{")
                    new_root = (tex_root / new_root).normpath()
                    sub_path_tex = new_root / sub_path_tex
                    if sub_path_tex.suffix == "":
                        sub_path_tex = sub_path_tex.with_suffix(".tex")
                    if not sub_path_tex.is_file():
                        status = FlattenStatus.FILE_NOT_FOUND
                else:
                    status = FlattenStatus.ILL_FORMATTED

            # Handle result
            if isinstance(sub_path_tex, str):
                inp_paths.append(sub_path_tex)
            if status != FlattenStatus.SUCCESS:
                if status == FlattenStatus.FILE_NOT_FOUND:
                    print(
                        f"Could not locate input file '{sub_path_tex}' "
                        f"on line {iline + 1} in '{path_tex}'",
                        file=sys.stderr,
                    )
                elif status == FlattenStatus.ILL_FORMATTED:
                    print(
                        f"Could not parse '{stripped}' on line {iline + 1} in '{path_tex}'",
                        file=sys.stderr,
                    )
                else:
                    print("Unknown error", file=sys.stderr)
                break
            if isinstance(sub_path_tex, str):
                status, sub_inp_paths = flatten_latex(sub_path_tex, fh_out, out_root, new_root)
                inp_paths.extend(sub_inp_paths)
                if status != FlattenStatus.SUCCESS:
                    break
            else:
                fh_out.write(rewrite_line(line, tex_root, out_root))
    return status, inp_paths


RE_REWRITE = re.compile(
    r"(?P<comopt>\\(?:includegraphics|thebibliography|verbatiminput)"
    r"(?:\s*\[.*?])?\s*)\{(?P<path>.*?)}"
)


def rewrite_line(line: str, tex_root: Path, out_root: str) -> str:
    r"""Rewrite the path in a source line.

    Parameters
    ----------
    line
        A line of LaTeX source code, possibly containing ``\includegraphics``
        or ``\thebibliography`` commands that need fixing.
    tex_root
        The directory with respect to which paths in the LaTeX source should
        be interpreted.
    out_root
        The new directory, with respect to which paths in the rewritten
        LaTeX source must be interpreted.

    Returns
    -------
    fixed_line
        A line in which the paths are fixed.
    """

    def repl(m):
        old_path = m.group("path")
        new_path = (tex_root / old_path).relpath(out_root).normpath()
        return m.group("comopt") + "{" + new_path + "}"

    return RE_REWRITE.sub(repl, line)


if __name__ == "__main__":
    main(sys.argv[1:])



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# © 2024–2025 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""RepRep wrapper for Typst.

This wrapper extracts relevant information from a Typst build
to inform StepUp of input files used or needed.

This is tested with Typst 0.14.
"""

import argparse
import contextlib
import shlex
import sys

from path import Path, TempDir

from stepup.core.api import amend, getenv
from stepup.core.utils import filter_dependencies
from stepup.core.worker import WorkThread

from .make_inventory import write_inventory


def main(argv: list[str] | None = None, work_thread: WorkThread | None = None):
    """Main program."""
    args = parse_args(argv)
    if work_thread is None:
        work_thread = WorkThread("stub")

    if not args.path_typ.endswith(".typ"):
        raise ValueError("The Typst source must have extension .typ")
    if not (args.path_out is None or args.path_out.suffix in (".pdf", ".png", ".svg")):
        raise ValueError("The Typst output must be a PDF, PNG, or SVG file.")

    # Get Typst executable and prepare some arguments that
    if args.typst is None:
        args.typst = getenv("REPREP_TYPST", "typst")

    # Prepare the command to run Typst
    typst_args = [args.typst, "compile", args.path_typ]
    if args.path_out is not None:
        typst_args.append(args.path_out)
    else:
        args.path_out = Path(args.path_typ[:-4] + ".pdf")
    if args.path_out.suffix == ".png":
        resolution = args.resolution
        if resolution is None:
            resolution = int(getenv("REPREP_TYPST_RESOLUTION", "144"))
        typst_args.append(f"--ppi={resolution}")
    for keyval in args.sysinp:
        typst_args.append("--input")
        typst_args.append(keyval)
    if len(args.typst_args) == 0:
        args.typst_args = shlex.split(getenv("REPREP_TYPST_ARGS", ""))
    typst_args.extend(args.typst_args)

    with contextlib.ExitStack() as stack:
        if args.keep_deps:
            # Remove any existing make-deps output from a previous run.
            path_dep = Path(args.path_typ[:-4] + ".dep")
            path_dep.remove_p()
        else:
            # Use a temporary file for the make-deps output.
            path_dep = stack.enter_context(TempDir()) / "typst.dep"
        typst_args.extend(["--deps", path_dep, "--deps-format", "make"])

        # Run typst compile
        returncode, stdout, stderr = work_thread.runsh(shlex.join(typst_args))
        print(stdout)
        # Get existing input files from the dependency file and amend.
        # Note that the deps file does not escape colons in paths,
        # so the code below assumes one never uses colons in paths.
        inp_paths = []
        if path_dep.is_file():
            out_paths = []
            with open(path_dep) as fh:
                dep_out, dep_inp = fh.read().split(":", 1)
                out_paths.extend(shlex.split(dep_out))
                inp_paths.extend(shlex.split(dep_inp))
        else:
            print(f"Dependency file not created: {path_dep}.", file=sys.stderr)
            out_paths = [args.path_out]

    # Look for missing input files in the standard error stream and amend them.
    if returncode != 0:
        lead = "error: file not found (searched at "
        inp_paths.extend(
            line[len(lead) : -1] for line in stderr.splitlines() if line.startswith(lead)
        )
    sys.stderr.write(stderr)
    inp_paths = filter_dependencies(inp_paths)
    amend(inp=inp_paths)

    # Write inventory
    if args.inventory is not None:
        inventory_paths = sorted(inp_paths) + out_paths
        write_inventory(args.inventory, inventory_paths, do_amend=False)

    # If the output path contains placeholders `{p}`, `{0p}`, or `{t}`,
    # we need to amend the output.
    if any(p in args.path_out for p in ("{p}", "{0p}", "{t}")):
        amend(out=out_paths)

    if returncode != 0:
        # Only use sys.exit in cases of an error,
        # so other programs may call this function without exiting.
        sys.exit(returncode)


def parse_args(argv: list[str] | None = None) -> argparse.Namespace:
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(
        prog="rr-compile-typst",
        description="Compile a Typst document and extract input and output info.",
    )
    parser.add_argument("path_typ", type=Path, help="The main typst source file.")
    parser.add_argument(
        "--out", dest="path_out", type=Path, help="The PDF/PNG/SVG output argument."
    )
    parser.add_argument(
        "--resolution",
        type=int,
        help="The resolution in DPI for PNG output. "
        "Defaults to ${REPREP_TYPST_RESOLUTION} or 144 if the variable is not defined.",
    )
    parser.add_argument(
        "--typst",
        help="The Typst executable. "
        "The default is ${REPREP_TYPST} or typst if the variable is not defined.",
    )
    parser.add_argument(
        "--keep-deps",
        help="Keep the dependency file after the compilation. "
        "The default is to use a temporary file, which is removed after it is processed.",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--inventory",
        type=Path,
        help="Write an inventory with all inputs and outputs, useful for archiving.",
    )
    parser.add_argument(
        "--sysinp",
        nargs="+",
        help="These key=value inputs to be passed to typst with `--input key=val."
        "Multiple key=value pairs are allowed after a single --sysinp.",
        default=(),
    )
    parser.add_argument(
        "typst_args",
        nargs="*",
        help="Additional arguments to be passed to typst. "
        "The defaults is `${REPREP_TYPST_ARGS}`, if the environment variable is defined.",
    )
    return parser.parse_args(argv)


if __name__ == "__main__":
    main(sys.argv[1:])



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# © 2024–2025 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Raster all pages in a PDF.

This is mainly intended to impede (the quality of) unauthorized copies.
"""

import argparse
import sys

import fitz

from stepup.core.api import getenv

__all__ = ("raster_pdf",)


def main(argv: list[str] | None = None):
    """Main program."""
    args = parse_args(argv)
    if args.resolution is None:
        args.resolution = int(getenv("REPREP_RASTER_RESOLUTION", "100"))
    if args.quality is None:
        args.quality = int(getenv("REPREP_RASTER_QUALITY", "50"))

    raster_pdf(args.path_inp, args.path_out, args.resolution, args.quality)
    return 0


def parse_args(argv: list[str] | None = None) -> argparse.Namespace:
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(prog="rr-raster-pdf", description="Raster a PDF file.")
    parser.add_argument("path_inp", help="The input PDF file.")
    parser.add_argument("path_out", help="The output PDF file.")
    parser.add_argument("-r", "--resolution", type=int, help="Bitmap resolution")
    parser.add_argument("-q", "--quality", type=int, help="JPEG quality")
    return parser.parse_args(argv)


def raster_pdf(path_inp: str, path_out: str, resolution: int, quality: int):
    """Convert a PDF into a rasterized version."""
    if not path_inp.endswith(".pdf"):
        raise ValueError(f"The input must have a `.pdf` extension, got: {path_inp}")
    if not path_out.endswith(".pdf"):
        raise ValueError(f"The output must have a `.pdf` extension, got: {path_out}")
    if resolution <= 0:
        raise ValueError(f"The resolution must be strictly positive, git: {resolution}")
    with fitz.open(path_inp) as src:
        dst = fitz.open()
        for src_page in src:
            src_page.wrap_contents()
            if src_page.rotation in (90, 270):
                height, width = src_page.mediabox_size
            else:
                width, height = src_page.mediabox_size
            pix = src_page.get_pixmap(dpi=resolution)
            stream = pix.tobytes(output="jpg", jpg_quality=quality)
            dst_page = dst.new_page(-1, width, height)
            dst_page.insert_image(dst_page.rect, stream=stream)
        dst.save(path_out, garbage=4, deflate=True, no_new_id=True)


if __name__ == "__main__":
    main(sys.argv[1:])



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# © 2024–2025 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Execution of notebooks with papermill.

The execution is implemented as a function here instead of calling
a `jupyter nbconvert` command directly, to avoid creating a subprocess.
Performing the execution in the worker process is more efficient,
especially when executing many notebooks that all require the same imports.
"""

import argparse
import json
import sys

from papermill import execute_notebook
from path import Path


def main(argv: list[str] | None = None):
    """Main program."""
    args = parse_args(argv)
    if not args.path_inp.endswith(".ipynb"):
        raise ValueError("The input must have a .ipynb extension.")
    if not args.path_out.endswith(".ipynb"):
        raise ValueError("The output must have a .ipynb extension.")
    execute_notebook(
        input_path=args.path_inp,
        output_path=args.path_out,
        parameters=json.loads(args.parameters) if args.parameters else {},
        progress_bar=False,
    )


def parse_args(argv: list[str] | None = None) -> argparse.Namespace:
    """Define command-line arguments."""
    parser = argparse.ArgumentParser(description="Convert a Jupyter notebook to HTML.")
    parser.add_argument(
        "path_inp", help="Path of the Jupyter notebook to execute and convert.", type=Path
    )
    parser.add_argument(
        "parameters", nargs="?", help="JSON serialized parameters for the notebook.", type=str
    )
    parser.add_argument("path_out", help="Path to the output notebook.", type=Path)
    return parser.parse_args(argv)


if __name__ == "__main__":
    main(sys.argv[1:])



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# © 2024–2025 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Application programming interface for StepUp RepRep."""

import json
import shlex
from collections.abc import Collection

from path import Path

from stepup.core.api import StepInfo, getenv, runsh, step, subs_env_vars
from stepup.core.utils import make_path_out, string_to_bool

__all__ = (
    "add_notes_pdf",
    "cat_pdf",
    "check_hrefs",
    "compile_latex",
    "compile_tectonic",
    "compile_typst",
    "convert_inkscape",
    "convert_inkscape_pdf",
    "convert_inkscape_png",
    "convert_jupyter",
    "convert_markdown",
    "convert_mutool",
    "convert_mutool_png",
    "convert_odf_pdf",
    "convert_weasyprint",
    "diff_latex",
    "execute_papermill",
    "flatten_latex",
    "make_inventory",
    "nup_pdf",
    "raster_pdf",
    "sanitize_bibtex",
    "sync_zenodo",
    "unplot",
    "wrap_git",
    "zip_inventory",
)


def add_notes_pdf(
    path_src: str, path_notes: str, path_dst: str, optional: bool = False, block: bool = False
) -> StepInfo:
    """Add a notes page at every even page of a PDF file.

    Parameters
    ----------
    path_src
        The original PDF document without notes pages.
    path_notes
        A single-page PDF document with a page suitable for taking notes.
    path_dst
        The output PDF with notes pages inserted.
    optional
        If `True`, the step is only executed when needed by other steps.
    block
        If `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    return step(
        "add-notes-pdf ${inp} ${out}",
        inp=[path_src, path_notes],
        out=path_dst,
        optional=optional,
        block=block,
    )


def cat_pdf(
    paths_inp: Collection[str],
    path_out: str,
    *,
    insert_blank: bool = False,
    optional: bool = False,
    block: bool = False,
) -> StepInfo:
    """Concatenate the pages of multiple PDFs into one document

    Parameters
    ----------
    paths_inp
        The input PDF files.
    path_out
        The concatenated PDF.
    insert_blank
        Insert a blank page after a PDF with an odd number of pages.
        The last page of each PDF is used to determine the size of the added blank page.
    optional
        If `True`, the step is only executed when needed by other steps.
    block
        If `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    args = ["cat-pdf", "${inp}", "${out}"]
    if insert_blank:
        args.append("--insert-blank")
    return step(
        " ".join(args),
        inp=paths_inp,
        out=path_out,
        optional=optional,
        block=block,
    )


def check_hrefs(path_src: str, path_config: str | None = None, block: bool = False) -> StepInfo:
    """Check hyper references in a Markdown, HTML or PDF file.

    Parameters
    ----------
    path_src
        The source Markdown, HTML or PDF to check.
    path_config
        The configuration file.
        Defaults to `${REPREP_CHECK_HREFS_CONFIG}` variable or `check_hrefs.yaml` if it is not set.
    block
        If `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    with subs_env_vars() as subs:
        path_src = subs(path_src)
        path_config = subs(path_config)
    args = ["check_hrefs", shlex.quote(path_src)]
    inp_paths = [path_src]
    if path_config is not None:
        inp_paths.append(path_config)
        args.extend(["-c", path_config])
    return step(" ".join(args), inp=inp_paths, block=block)


def compile_latex(
    path_tex: str,
    *,
    run_bibtex=True,
    maxrep: int = 5,
    workdir: str = "./",
    latex: str | None = None,
    bibtex: str | None = None,
    inventory: str | bool | None = None,
    optional: bool = False,
    block: bool = False,
) -> StepInfo:
    """Create a step for the compilation of a LaTeX source.

    Parameters
    ----------
    path_tex
        The main tex source file.
        This argument may contain environment variables.
    run_bibtex
        By default, when bib files are used, BibTeX is invoked.
        This can be overruled by setting this argument to False,
        which is useful when recompiling sources with fixed bbl files.
    maxrep
        The maximum number of repetitions of the LaTeX command
        in case the aux file keeps changing.
    workdir
        The working directory where the LaTeX command must be executed.
    latex
        Path to the LaTeX executable. Note that only PDF-producing LaTeX compilers are supported:
        `pdflatex`, `xelatex` or `lualatex`.
        Defaults to `${REPREP_LATEX}` variable or `pdflatex` if the variable is unset.
    bibtex
        Path to the BibTeX executable.
        Defaults to `${REPREP_BIBTEX}` variable or `bibtex` if the variable is unset.
    inventory
        If set to a `str`, it specifies the inventory file to write.
        If set to  `True`, the inventory file is written to the default location,
        which is the stem of the source file with `-inventory.txt` appended.
        When the environment variable `REPREP_LATEX_INVENTORY` is set to `1`,
        the inventory file is always written, unless this argument is set to `False`.
    optional
        If `True`, the step is only executed when needed by other steps.
    block
        If `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.

    Notes
    -----
    The LaTeX source is compiled with the `rr-compile-latex` command,
    which can detect dependencies on other files by scanning for
    `\\input`, `\\include`, `\\includegraphics`, etc.
    Due to the complexity of LaTeX, the dependency scanning is not perfect.
    You can manually specify dependencies in the LaTeX source with the following comments:

    -  `%REPREP inp inp_path`
    -  `%REPREP out out_path`
    -  `%REPREP vol vol_path`

    When any of these paths is a relative path,
    it is interpreted in the same way as the LaTeX compiler would resolve it.
    You can also hide lines from the dependency scanner by adding `%REPREP ignore`
    at the end of a line.
    """
    with subs_env_vars() as subs:
        path_tex = subs(path_tex)
    if not path_tex.endswith(".tex"):
        raise ValueError(f"The input of the latex command must end with .tex, got {path_tex}.")

    prefix = path_tex[:-4]
    path_pdf = f"{prefix}.pdf"

    args = ["compile-latex", shlex.quote(path_tex)]
    inp_paths = [path_tex]
    out_paths = [path_pdf, f"{prefix}.aux", f"{prefix}.fls"]
    if maxrep != 5:
        args.append("--maxrep=" + shlex.quote(str(maxrep)))
    if latex is not None:
        args.append("--latex=" + shlex.quote(latex))
    if run_bibtex:
        args.append("--run-bibtex")
        if bibtex is not None:
            args.append("--bibtex=" + shlex.quote(bibtex))
    if inventory is None:
        inventory = string_to_bool(getenv("REPREP_LATEX_INVENTORY", "0"))
    if inventory is True:
        inventory = f"{prefix}-inventory.txt"
    if isinstance(inventory, str):
        args.append("--inventory=" + shlex.quote(inventory))
        out_paths.append(inventory)
    return step(
        " ".join(args),
        inp=inp_paths,
        out=out_paths,
        workdir=workdir,
        optional=optional,
        block=block,
    )


def compile_tectonic(
    path_tex: str,
    dest: str | None = None,
    *,
    workdir: str = "./",
    tectonic: str | None = None,
    keep_deps: bool = False,
    tectonic_args: Collection[str] = (),
    inventory: str | bool | None = None,
    optional: bool = False,
    block: bool = False,
) -> StepInfo:
    """Create a step for the compilation of a LaTeX source with Tectonic.

    !!! warning

        This feature is only tested with Tectonic 0.15

        Support for Tectonic in StepUp RepRep is experimental.
        Expect breaking changes in future releases.

    Parameters
    ----------
    path_tex
        The main LaTeX source file.
        This argument may contain environment variables.
    dest
        Output destination: `None`, a directory or a file.
    workdir
        The working directory where the LaTeX command must be executed.
    tectonic
        Path to the Tectonic executable.
        Defaults to `${REPREP_TECTONIC}` variable or `tectonic` if the variable is unset.
    keep_deps
        If `True`, the dependency file is kept after the compilation.
        The dependency file is also kept if the environment variable
        `REPREP_KEEP_TECTONIC_DEPS` is set to `"1"`.
    tectonic_args
        Additional arguments for tectonic.
        The defaults is `${REPREP_TECTONIC_ARGS}`, if the environment variable is defined.
    inventory
        If set to a `str`, it specifies the inventory file to write.
        If set to  `True`, the inventory file is written to the default location,
        which is the stem of the source file with `-inventory.txt` appended.
        When the environment variable `REPREP_TECTONIC_INVENTORY` is set to `1`,
        the inventory file is always written, unless this argument is set to `False`.
    optional
        If `True`, the step is only executed when needed by other steps.
    block
        If `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    with subs_env_vars() as subs:
        path_tex = subs(path_tex)
        dest = subs(dest)
    if not path_tex.endswith(".tex"):
        raise ValueError(f"The input of the tectonic command must end with .tex, got {path_tex}.")
    path_out = make_path_out(path_tex, dest, ".pdf")

    stem = path_tex[:-4]
    args = ["compile-tectonic"]
    if tectonic is not None:
        args.append(f"--tectonic={shlex.quote(tectonic)}")
    paths_out = [path_out]
    if keep_deps or string_to_bool(getenv("REPREP_KEEP_TECTONIC_DEPS", "0")):
        args.append("--keep-deps")
        paths_out.append(f"{stem}.dep")
    if inventory is None:
        inventory = string_to_bool(getenv("REPREP_TECTONIC_INVENTORY", "0"))
    if inventory is True:
        inventory = f"{stem}-inventory.txt"
    if isinstance(inventory, str):
        args.append(f"--inventory={shlex.quote(inventory)}")
        paths_out.append(inventory)
    args.append(shlex.quote(path_tex))
    if path_tex[:-4] != path_out[:-4]:
        args.append("--out=" + shlex.quote(path_out))
    path_inp = [path_tex]
    if len(tectonic_args) > 0:
        args.append("--")
        args.extend(shlex.quote(tectonic_arg) for tectonic_arg in tectonic_args)

    return step(
        " ".join(args),
        inp=path_inp,
        out=paths_out,
        workdir=workdir,
        optional=optional,
        block=block,
    )


def compile_typst(
    path_typ: str,
    dest: str | None = None,
    *,
    sysinp: dict[str, str | Path] | None = None,
    resolution: int | None = None,
    workdir: str = "./",
    typst: str | None = None,
    keep_deps: bool = False,
    typst_args: Collection[str] = (),
    inventory: str | bool | None = None,
    optional: bool = False,
    block: bool = False,
) -> StepInfo:
    """Create a step for the compilation of a Typst source.

    !!! warning

        This feature will only work well with Typst 0.14 or later.

        Support for Typst in StepUp RepRep is experimental.
        Expect breaking changes in future releases.
        Some limitations include:

        - SVG figures with references to external SVG images are not processed correctly.
          These images are not rendered, neither are they included in the dep file.
          This currently being addressed in the following issue:
          https://github.com/typst/typst/issues/6858
        - When the typst compiler detects an error in the input, it doesn't write the dep file.
          While this is the desirable behavior for Make-like tools, it does not work well in StepUp.
          This issue is fixed in the main branch of typst, but not yet in a released version:
          https://github.com/typst/typst/pull/7209
          After the next release of Typst, StepUp RepRep will be updated to use of JSON dep files:
          https://github.com/reproducible-reporting/stepup-reprep/pull/22

    Parameters
    ----------
    path_typ
        The main typst source file.
        This argument may contain environment variables.
    dest
        Output destination: `None`, a directory or a file.
        For SVG and PNG outputs, this argument must be specified with the desired extension.
        If the output contains any of `{p}`, `{0p}` or `{t}`, the output paths are not
        known a priori and will be amended.
    sysinp
        A dictionary with the input arguments passed to `typst` with `--input key=val`.
        Items are ignored when keys are not strings or when values
        are not of type `str`, `int`, `float` or `Path`.
        If a key is not a valid Python identifier, an exception is raised.
        When values are `Path` instances, they are treated as input dependencies for the step.
        These parameters are available in the document as `#sys.inputs.key`.
        One may also provide an object, which is converted into a `dict` with the `vars()` built-in.
    resolution
        The resolution of the bitmap in dots per inch (dpi),
        only relevant for PNG output.
    workdir
        The working directory where the LaTeX command must be executed.
    typst
        Path to the Typst executable.
        Defaults to `${REPREP_TYPST}` variable or `typst` if the variable is unset.
    keep_deps
        If `True`, the dependency file is kept after the compilation.
        The dependency file is also kept if the environment variable
        `REPREP_KEEP_TYPST_DEPS` is set to `"1"`.
    typst_args
        Additional arguments for typst.
        The defaults is `${REPREP_TYPST_ARGS}`, if the environment variable is defined.
    inventory
        If set to a `str`, it specifies the inventory file to write.
        If set to  `True`, the inventory file is written to the default location,
        which is the stem of the source file with `-inventory.txt` appended.
        When the environment variable `REPREP_TYPST_INVENTORY` is set to `1`,
        the inventory file is always written, unless this argument is set to `False`.
    optional
        If `True`, the step is only executed when needed by other steps.
    block
        If `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    with subs_env_vars() as subs:
        path_typ = subs(path_typ)
        dest = subs(dest)
    if not path_typ.endswith(".typ"):
        raise ValueError(f"The input of the typst command must end with .typ, got {path_typ}.")
    path_out = make_path_out(path_typ, dest, ".pdf", [".svg", ".png"])

    stem = path_typ[:-4]
    args = ["compile-typst"]
    if resolution is not None:
        args.append(f"--resolution={shlex.quote(str(resolution))}")
    if typst is not None:
        args.append(f"--typst={shlex.quote(typst)}")
    paths_out = []
    if not any(x in path_out for x in ("{p}", "{0p}", "{t}")):
        paths_out.append(path_out)
    if keep_deps or string_to_bool(getenv("REPREP_KEEP_TYPST_DEPS", "0")):
        args.append("--keep-deps")
        paths_out.append(f"{stem}.dep")
    if inventory is None:
        inventory = string_to_bool(getenv("REPREP_TYPST_INVENTORY", "0"))
    if inventory is True:
        inventory = f"{stem}-inventory.txt"
    if isinstance(inventory, str):
        args.append(f"--inventory={shlex.quote(inventory)}")
        paths_out.append(inventory)
    args.append(shlex.quote(path_typ))
    if path_typ[:-4] != path_out[:-4]:
        args.append("--out=" + shlex.quote(path_out))
    path_inp = [path_typ]
    if sysinp is not None:
        if not isinstance(sysinp, dict):
            sysinp = vars(sysinp)
        if len(sysinp) > 0:
            args.append("--sysinp")
            for key, val in sysinp.items():
                if not isinstance(key, str):
                    continue
                if not isinstance(val, str | int | float | Path):
                    continue
                if not key.isidentifier():
                    raise ValueError(f"Invalid sysinp key: {key}")
                args.append(shlex.quote(str(key)) + "=" + shlex.quote(str(val)))
                if isinstance(val, Path):
                    path_inp.append(val)
    if len(typst_args) > 0:
        args.append("--")
        args.extend(shlex.quote(typst_arg) for typst_arg in typst_args)

    return step(
        " ".join(args),
        inp=path_inp,
        out=paths_out,
        workdir=workdir,
        optional=optional,
        block=block,
    )


def convert_inkscape(
    path_svg: str,
    path_out: str,
    *,
    inkscape: str | None = None,
    inkscape_args: Collection[str] = (),
    optional: bool = False,
    block: bool = False,
) -> StepInfo:
    """Convert an SVG figure to a PDF file, detecting dependencies of the SVG on other files.

    Parameters
    ----------
    path_svg
        The input SVG figure.
        It may contain <img> tags referring to other files included in the figure.
    path_out
        The output PDF or PNG file. Other formats are not supported.
    inkscape
        The path to the inkscape executable.
        Defaults to `${REPREP_INKSCAPE}` variable or `inkscape` if the variable is unset.
    inkscape_args
        Additional arguments to pass to inkscape. E.g. `["-T"]` to convert text to glyphs in PDFs.
        Depending on the extension of the output, the default is `${REPREP_INKSCAPE_PDF_ARGS}` or
        `${REPREP_INKSCAPE_PNG_ARGS}`, if the environment variable is defined.
    optional
        If `True`, the step is only executed when needed by other steps.
    block
        If `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.

    Notes
    -----
    A wrapper around inkscape is used to carry out the conversion: `stepup.reprep.convert_inkscape`.
    The wrapper scans the SVG for dependencies, which may be a bit slow in case of large files.
    """
    with subs_env_vars() as subs:
        path_svg = subs(path_svg)
        path_out = subs(path_out)
    if not path_svg.endswith(".svg"):
        raise ValueError("The SVG file must have extension .svg")
    if not path_out.endswith((".pdf", ".png")):
        raise ValueError("The output file must have extension .pdf or .png")
    args = ["convert-inkscape", shlex.quote(path_svg), shlex.quote(path_out)]
    if inkscape is not None:
        args.append("--inkscape=" + shlex.quote(inkscape))
    if len(inkscape_args) > 0:
        args.append("--")
        args.extend(shlex.quote(inkscape_arg) for inkscape_arg in inkscape_args)
    return step(" ".join(args), inp=path_svg, out=path_out, block=block, optional=optional)


def convert_inkscape_pdf(
    path_svg: str,
    dest: str | None = None,
    *,
    inkscape: str | None = None,
    inkscape_args: Collection[str] = (),
    optional: bool = False,
    block: bool = False,
) -> StepInfo:
    """Shorthand for `convert_inkscape` with the output file derived from the SVG file.

    The `dest` argument can be `None`, a directory or a file.
    """
    with subs_env_vars() as subs:
        path_svg = subs(path_svg)
        dest = subs(dest)
    if not path_svg.endswith(".svg"):
        raise ValueError("The SVG file must have extension .svg")
    path_pdf = make_path_out(path_svg, dest, ".pdf")
    return convert_inkscape(
        path_svg,
        path_pdf,
        inkscape=inkscape,
        inkscape_args=inkscape_args,
        optional=optional,
        block=block,
    )


def convert_inkscape_png(
    path_svg: str,
    dest: str | None = None,
    *,
    inkscape: str | None = None,
    inkscape_args: Collection[str] = (),
    optional: bool = False,
    block: bool = False,
) -> StepInfo:
    """Shorthand for `convert_inkscape` with the output file derived from the SVG file.

    The `dest` argument can be `None`, a directory or a file. See `make_path_out`.
    """
    with subs_env_vars() as subs:
        path_svg = subs(path_svg)
        dest = subs(dest)
    if not path_svg.endswith(".svg"):
        raise ValueError("The SVG file must have extension .svg")
    path_png = make_path_out(path_svg, dest, ".png")
    return convert_inkscape(
        path_svg,
        path_png,
        inkscape=inkscape,
        inkscape_args=inkscape_args,
        optional=optional,
        block=block,
    )


def convert_jupyter(
    path_nb: str,
    dest: str | None = None,
    *,
    inp: str | Collection[str] = (),
    out: str | Collection[str] = (),
    execute: bool = True,
    to: str | None = None,
    nbargs: str | dict | list | None = None,
    jupyter: str | None = None,
    optional: bool = False,
    pool: str | None = None,
    block: bool = False,
) -> StepInfo:
    """Convert a Jupyter notebook, by default to HTML with execution of cells.

    !!! warning

        - Support for `juptyer nbconvert` in StepUp RepRep is experimental.
          Expect breaking changes in future releases.
        - The conversion uses `nbconvert`, which has an unfixed security vulnerability.
          See https://access.redhat.com/security/cve/cve-2025-53000

    Parameters
    ----------
    path_nb
        The input Jupyter notebook.
    dest
        Output destination: `None`, a directory or a file.
    inp
        One or more input files used by the notebook.
        You can also declare inputs with `amend(inp=...)` in the notebook,
        but specifying them here will make the scheduling more efficient.
    out
        One or more output files produced by the notebook.
        You can also declare outputs with `amend(out=...)` in the notebook,
        but you can specify them here if you want to make the notebook execution optional,
        i.e. dependent on whether the outputs are used in other steps.
    execute
        If `True`, the notebook is executed before conversion.
    to
        The output format. The default depends on the extension of the output file.
        if `to` is given and `dest` is `None` or a directory,
        the `to` argument is used to determine the output file extension.
    nbargs
        If `str`, it is passed literally as additional argument to the notebook
        through the environment variable `REPREP_NBARGS`.
        If `dict` or `list`, it is converted to a JSON string first.
        The notebook should read this variable with `os.getenv("REPREP_NBARGS")`
        and not `stepup.core.api.getenv()` because the variable is local to the process.
        It is impossible (and pointless) for the StepUp director to detect changes in this variable.
        Even if it is globally defined, it will be overridden in this step.
    jupyter
        The path to the jupyter executable.
        Defaults to `${REPREP_JUPYTER}` variable or `jupyter` if the variable is unset.
    optional
        If `True`, the step is only executed when needed by other steps.
    pool
        The pool in which the step is executed,
        which may be convenient to limit the number of parallel notebooks being executed,
        e.g. when the already run calculations in parallel.
    block
        If `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    with subs_env_vars() as subs:
        path_nb = subs(path_nb)
        dest = subs(dest)
    if not path_nb.endswith(".ipynb"):
        raise ValueError("The notebook file must have extension .ipynb")
    if isinstance(inp, str):
        inp = [inp]
    if isinstance(out, str):
        out = [out]
    default_exts = {
        "html": ".html",
        "pdf": ".pdf",
        "notebook": ".ipynb",
        "latex": ".tex",
        "markdown": ".md",
        "rst": ".rst",
        "script": ".py",
        "asciidoc": ".txt",
    }
    if to is not None:
        default_ext = default_exts.get(to)
        if default_ext is None:
            raise ValueError(f"Unsupported output format: {to}")
    else:
        default_ext = ".html"
    other_exts = [".html", ".pdf", ".ipynb", ".tex", ".md", ".rst", ".py", ".txt"]
    path_out = make_path_out(path_nb, dest, default_ext, other_exts)
    if to is None:
        default_formats = {val: key for key, val in default_exts.items()}
        to = default_formats[path_out.suffix]
    if jupyter is None:
        jupyter = getenv("REPREP_JUPYTER", "jupyter")
    args = [jupyter, "nbconvert", shlex.quote(path_nb), "--stdout", "--to", to]
    if execute:
        args.append("--execute")
    if nbargs is not None:
        if isinstance(nbargs, dict | list):
            nbargs = json.dumps(nbargs)
        elif not isinstance(nbargs, str):
            nbargs = str(nbargs)
        args.insert(0, "REPREP_NBARGS=" + shlex.quote(nbargs))
    args.extend([">", shlex.quote(path_out)])
    runsh(
        " ".join(args),
        inp=[path_nb, *inp],
        out=[path_out, *out],
        optional=optional,
        pool=pool,
        block=block,
    )


def convert_markdown(
    path_md: str,
    dest: str | None = None,
    *,
    paths_css: str | Collection[str] = (),
    optional: bool = False,
    block: bool = False,
) -> StepInfo:
    """Convert a markdown to HTML.

    Parameters
    ----------
    path_md
        The markdown input file.
    dest
        Output destination: `None`, a directory or a file.
    paths_css
        Path or multiple paths of a local CSS file, or a list of multiple such paths,
        to be included in the HTML header.
        Note that one may also specify CSS file in the markdown header.
        Defaults to `${REPREP_MARKDOWN_CSS}` if the variable is set,
        which is interpreted as a colon-separated list of files.
    optional
        If `True`, the step is only executed when needed by other steps.
    block
        If `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    with subs_env_vars() as subs:
        path_md = subs(path_md)
        dest = subs(dest)
    if not path_md.endswith(".md"):
        raise ValueError("The Markdown file must have extension .md")
    path_html = make_path_out(path_md, dest, ".html")
    inp = [path_md]
    args = ["convert-markdown", shlex.quote(path_md), shlex.quote(path_html)]
    if len(paths_css) > 0:
        if isinstance(paths_css, str):
            paths_css = [paths_css]
        args.append("--css")
        args.extend(shlex.quote(path_css) for path_css in paths_css)
        inp.extend(paths_css)
    return step(" ".join(args), inp=inp, out=path_html, optional=optional, block=block)


def convert_mutool(
    path_pdf: str,
    path_out: str,
    *,
    resolution: int | None = None,
    mutool: str | None = None,
    optional: bool = False,
    block: bool = False,
) -> StepInfo:
    """Convert a PDF to a bitmap with mutool (from MuPDF).

    Parameters
    ----------
    path_pdf
        The input PDF file.
    path_out
        The output image file.
    resolution
        The resolution of the output bitmap in dots per inch (dpi).
    mutool
        The path to the mutool executable.
        Defaults to `${REPREP_MUTOOL}` variable or `mutool` if the variable is unset.
    optional
        If `True`, the step is only executed when needed by other steps.
    block
        If `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    if resolution is None:
        resolution = int(getenv("REPREP_CONVERT_PDF_RESOLUTION", "100"))
    if mutool is None:
        mutool = getenv("REPREP_MUTOOL", "mutool")
    args = [shlex.quote(mutool), "draw -q -o ${out} -r", shlex.quote(str(resolution)), "${inp}"]
    return runsh(
        " ".join(args),
        inp=path_pdf,
        out=path_out,
        optional=optional,
        block=block,
    )


def convert_mutool_png(
    path_pdf: str,
    dest: str | None = None,
    *,
    resolution: int | None = None,
    mutool: str | None = None,
    optional: bool = False,
    block: bool = False,
) -> StepInfo:
    """Shorthand for `convert_mutool` with the output file derived from the PDF file.

    The `dest` argument can be `None`, a directory or a file. See `make_path_out`.
    """
    with subs_env_vars() as subs:
        path_pdf = subs(path_pdf)
        dest = subs(dest)
    if not path_pdf.endswith(".pdf"):
        raise ValueError("The PDF file must have extension .pdf")
    path_png = make_path_out(path_pdf, dest, ".png")
    return convert_mutool(
        path_pdf, path_png, resolution=resolution, mutool=mutool, optional=optional, block=block
    )


def convert_weasyprint(
    path_html: str,
    dest: str | None = None,
    *,
    weasyprint: str | None = None,
    optional: bool = False,
    block: bool = False,
) -> StepInfo:
    """Convert a HTML document to PDF.

    Parameters
    ----------
    path_html
        The HTML input file.
    dest
        Output destination: `None`, a directory or a file.
    weasyprint
        The path to the weasyprint executable.
        Defaults to `${REPREP_WEASYPRINT}` variable or `weasyprint` if the variable is unset.
    optional
        If `True`, the step is only executed when needed by other steps.
    block
        If `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    with subs_env_vars() as subs:
        path_html = subs(path_html)
        dest = subs(dest)
    if not path_html.endswith(".html"):
        raise ValueError("The HTML file must have extension .html")
    path_pdf = make_path_out(path_html, dest, ".pdf")
    args = [
        "convert-weasyprint",
        shlex.quote(path_html),
        shlex.quote(path_pdf),
    ]
    if weasyprint is not None:
        args.append("--weasyprint=" + shlex.quote(weasyprint))
    return step(" ".join(args), inp=path_html, out=path_pdf, block=block, optional=optional)


def convert_odf_pdf(
    path_odf: str,
    dest: str | None = None,
    *,
    libreoffice: str | None = None,
    optional: bool = False,
    block: bool = False,
) -> StepInfo:
    """Convert a file in OpenDocument format to PDF.

    Parameters
    ----------
    path_odf
        The input open-document file.
    dest
        None, output directory or path. See `make_path_out`.
    libreoffice
        The libreoffice executable.
        Defaults to `${REPREP_LIBREOFFICE}` variable or `libreoffice` if the variable is unset.
    optional
        If `True`, the step is only executed when needed by other steps.
    block
        If `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.

    Notes
    -----
    This function does not yet scan the source document for reference to external files.
    which should ideally be added as dependencies.
    """
    with subs_env_vars() as subs:
        path_odf = subs(path_odf)
        dest = subs(dest)
    if libreoffice is None:
        libreoffice = getenv("REPREP_LIBREOFFICE", "libreoffice")
    command = (
        # Simple things should be simple! ;) See:
        # https://bugs.documentfoundation.org/show_bug.cgi?id=106134
        # https://bugs.documentfoundation.org/show_bug.cgi?id=152192
        # Not solved yet:
        # https://bugs.documentfoundation.org/show_bug.cgi?id=160033
        "WORK=`mktemp -d --suffix=reprep` && "
        + shlex.quote(libreoffice)
        + " -env:UserInstallation=file://${WORK} --convert-to pdf ${inp} --outdir ${WORK} "
        "> /dev/null && cp ${WORK}/*.pdf ${out} && rm -r ${WORK}"
    )
    path_pdf = make_path_out(path_odf, dest, ".pdf")
    return runsh(command, inp=path_odf, out=path_pdf, optional=optional, block=block)


DEFAULT_LATEXDIFF_ARGS = (
    "--append-context2cmd=abstract,supplementary,dataavailability,funding,"
    "authorcontributions,conflictsofinterest,abbreviations",
)


def diff_latex(
    path_old: str,
    path_new: str,
    path_diff: str,
    *,
    latexdiff: str | None = None,
    latexdiff_args: Collection[str] = DEFAULT_LATEXDIFF_ARGS,
    optional: bool = False,
    block: bool = False,
) -> StepInfo:
    r"""Create a step to run latexdiff.

    Parameters
    ----------
    path_old
        The old tex or bbl source.
    path_new
        The new tex or bbl source.
    path_diff
        The diff output tex or bbl.
    latexdiff
        Path of the latexdiff  executable.
        Defaults to `${REPREP_LATEXDIFF}` variable or `latexdiff` if the variable is unset.
    latexdiff_args
        Additional arguments for latexdiff.
        Defaults to `${REPREP_LATEXDIFF_ARG}` variable.
        If this variable is unset, the following default is used:

        ```
        --append-context2cmd=abstract,supplementary,dataavailability,funding, \
                             authorcontributions,conflictsofinterest,abbreviations
        ```

        The option `--no-label` is always added because it is needed to make the file reproducible.
    optional
        If `True`, the step is only executed when needed by other steps.
    block
        If `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    if latexdiff is None:
        latexdiff = getenv("REPREP_LATEXDIFF", "latexdiff")

    if latexdiff_args is None:
        latexdiff_args = shlex.split(getenv("REPREP_LATEXDIFF_ARGS", ""))

    args = [shlex.quote(latexdiff)]
    args.extend(shlex.quote(latexdiff_arg) for latexdiff_arg in latexdiff_args)
    args.extend(["${inp}", "--no-label", ">", "${out}"])
    return runsh(
        " ".join(args),
        inp=[path_old, path_new],
        out=path_diff,
        optional=optional,
        block=block,
    )


def execute_papermill(
    path_nb: str,
    dest: str,
    *,
    inp: str | Collection[str] = (),
    out: str | Collection[str] = (),
    parameters: dict[str] | None = None,
    optional: bool = False,
    pool: str | None = None,
    block: bool = False,
) -> StepInfo:
    """Execute a Jupyter Notebook with papermill and save the notebook with outputs as a new file.

    Parameters
    ----------
    path_nb
        The input Jupyter notebook.
    dest
        Output destination: a directory or a file.
        The output format is always IPYNB.
    inp
        One or more input files used by the notebook.
        You can also declare inputs with `amend(inp=...)` in the notebook,
        but specifying them here will make the scheduling more efficient.
    out
        One or more output files produced by the notebook.
        You can also declare outputs with `amend(out=...)` in the notebook,
        but you can specify them here if you want to make the notebook execution optional,
        i.e. dependent on whether the outputs are used in other steps.
    parameters
        The dictionary, if any, must be JSON-serializable.
        It will be passed to the parameters argument of `papermill.execute_notebook()`.
    optional
        If `True`, the step is only executed when needed by other steps.
    pool
        The pool in which the step is executed,
        which may be convenient to limit the number of parallel notebooks being executed,
        e.g. when the already run calculations in parallel.
    block
        If `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    with subs_env_vars() as subs:
        path_nb = subs(path_nb)
        dest = subs(dest)
    if not path_nb.endswith(".ipynb"):
        raise ValueError("The notebook file must have extension .ipynb")
    if isinstance(inp, str):
        inp = [inp]
    if isinstance(out, str):
        out = [out]
    path_out = make_path_out(path_nb, dest, ".ipynb")
    if parameters is None:
        parameters = {}
    args = ["execute-papermill", shlex.quote(path_nb)]
    if len(parameters) > 0:
        args.append(shlex.quote(json.dumps(parameters)))
    args.append(shlex.quote(path_out))
    step(
        " ".join(args),
        inp=[path_nb, *inp],
        out=[path_out, *out],
        optional=optional,
        pool=pool,
        block=block,
    )


def flatten_latex(path_tex: str, path_flat: str, *, optional: bool = False, block: bool = False):
    r"""Flatten structured LaTeX source files (substitute `\input` and friends by their content).

    Parameters
    ----------
    path_tex
        The main tex file to be converted.
    path_flat
        The flattened output file.
    optional
        If `True`, the step is only executed when needed by other steps.
    block
        If `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    return step(
        "flatten-latex ${inp} ${out}",
        inp=path_tex,
        out=path_flat,
        optional=optional,
        block=block,
    )


def make_inventory(
    *paths: Collection[str],
    path_def: str | None = None,
    optional: bool = False,
    block: bool = False,
) -> StepInfo:
    """Create an `inventory.txt` file.

    Parameters
    ----------
    paths
        Paths to include in the `inventory.txt` file,
        except for the last, which is the inventory file to write.
    path_def
        An inventory definitions file, used to constructe the list of files.
    optional
        If `True`, the step is only executed when needed by other steps.
    block
        If `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    if len(paths) < 1:
        raise ValueError("At least one path must be given.")
    paths_inp = list(paths[:-1])
    args = ["make-inventory", *paths_inp]
    if path_def is not None:
        args.extend(["-i", path_def])
        paths_inp.append(path_def)
    args.extend(["-o", paths[-1]])
    return step(
        shlex.join(args),
        inp=paths_inp,
        out=[paths[-1]],
        optional=optional,
        block=block,
    )


def nup_pdf(
    path_src: str,
    path_dst: str,
    *,
    nrow: int | None = None,
    ncol: int | None = None,
    margin: float | None = None,
    page_format: str | None = None,
    optional: bool = False,
    block: bool = False,
) -> StepInfo:
    """Put multiple pages per sheet using a fixed layout.

    Parameters
    ----------
    path_src
        The original PDF document (with normal pages).
    path_dst
        The output PDF with (multiple pages per sheet).
    nrow
        The number of rows on each output sheet.
        The default is `${REPREP_NUP_NROW}` or 2 if the variable is not set.
    ncol
        The number of columns on each output sheet.
        The default is `${REPREP_NUP_NCOL}` or 2 if the variable is not set.
    margin
        The margin in mm between the pages on each sheet. (Also used as sheet margin.)
        The default is `${REPREP_NUP_MARGIN}` or 10.0 if the variable is not set.
    page_format
        The output page format
        The default is `${REPREP_NUP_PAGE_FORMAT}` or A4-L if the variable is not set.
    optional
        If `True`, the step is only executed when needed by other steps.
    block
        If `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    args = ["nup-pdf", "${inp}", "${out}"]
    if nrow is not None:
        args.extend(["-r", str(nrow)])
    if ncol is not None:
        args.extend(["-c", str(ncol)])
    if margin is not None:
        args.extend(["-m", str(margin)])
    if page_format is not None:
        args.extend(["-p", shlex.quote(page_format)])
    return step(" ".join(args), inp=path_src, out=path_dst, optional=optional, block=block)


def raster_pdf(
    path_inp: str,
    dest: str,
    *,
    resolution: int | None = None,
    quality: int | None = None,
    optional: bool = False,
    block: bool = False,
) -> StepInfo:
    """Turn each page of a PDF into a rendered JPEG bitmap contained in a new PDF.

    Parameters
    ----------
    path_inp
        The input PDF file.
    dest
        None, output directory or path. See `make_path_out`.
    resolution
        The resolution of the bitmap in dots per inch (dpi).
        The default value is taken from `${REPREP_RASTER_RESOLUTION}` or 100 if the variable is not
        set.
    quality
        The JPEG quality of the bitmap.
        The default value is taken from `${REPREP_RASTER_QUALITY}` or 50 if the variable is not set.
    optional
        If `True`, the step is only executed when needed by other steps.
    block
        If `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    args = ["raster-pdf", "${inp}", "${out}"]
    if resolution is not None:
        args.extend(["-r", str(resolution)])
    if quality is not None:
        args.extend(["-q", shlex.quote(str(quality))])
    path_out = make_path_out(path_inp, dest, ".pdf")
    return step(" ".join(args), inp=path_inp, out=path_out, optional=optional, block=block)


def sanitize_bibtex(
    path_bib: str,
    *,
    path_aux: str | None = None,
    path_cfg: str | None = None,
    path_out: str | None = None,
    overwrite: bool = False,
    optional: bool = False,
    block: bool = False,
) -> StepInfo:
    """Sanitize a BibTeX file.

    Parameters
    ----------
    path_bib
        Path to the BibTeX file to be sanitized.
    path_aux
        Paths to LaTeX aux file if any.
        This is used to detect unused or missing citations.
    path_cfg
        The YAML configuration file for the `rr-sanitize-bibtex` script.
    path_out
        If given, a single cleaned-up bibtex file is written as output,
        which you can manually copy back to the original if you approve of the cleanup.
        If not given, the original bibtex file is overwritten (if there is only one),
        which will drain the scheduler.
        You then check if the updated version is correct and rerun the build to approve.
    overwrite
        If `True`, it is assumed that you want to overwrite an input bibtex file,
        in which case `path_out` is not treated as a new output file in the workflow.
    optional
        If `True`, the step is only executed when needed by other steps.
    block
        If `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    with subs_env_vars() as subs:
        path_bib = subs(path_bib)
        path_cfg = subs(path_cfg)
        path_aux = subs(path_aux)
        path_out = subs(path_out)
    if not path_bib.endswith(".bib"):
        raise ValueError("The BibTeX file must have extension .bib")

    args = ["bibsane", shlex.quote(path_bib)]
    paths_inp = [path_bib]
    if path_aux is not None:
        args.append("--aux=" + shlex.quote(path_aux))
        paths_inp.append(path_aux)
    if path_cfg is not None:
        args.append("--config=" + shlex.quote(path_cfg))
        paths_inp.append(path_cfg)
    paths_out = []
    if path_out is not None:
        args.append("--out=" + shlex.quote(path_out))
        if not overwrite:
            paths_out.append(path_out)
    return step(" ".join(args), inp=paths_inp, out=paths_out, optional=optional, block=block)


def sync_zenodo(path_config: str, *, verbose: bool = False, block: bool = False) -> StepInfo:
    """Synchronize data with an draft dataset on Zenodo.

    Parameters
    ----------
    path_config
        The YAML configuration file for the Zenodo upload.
    verbose
        Set to True to print Zenodo API requests and responses to the standard output.
    block
        If `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    command = "sync-zenodo ${inp}"
    if verbose:
        command += " --verbose"
    return step(command, inp=path_config, block=block)


def unplot(
    path_svg: str, dest: str | None = None, *, optional: bool = False, block: bool = False
) -> StepInfo:
    """Convert a plot back to data.

    Parameters
    ----------
    path_svg
        The SVG file with paths to be converted back.
    dest
        An output directory or file.

    optional
        If `True`, the step is only executed when needed by other steps.
    block
        If `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    path_out = make_path_out(path_svg, dest, ".json")
    command = "unplot ${inp} ${out}"
    return step(command, inp=path_svg, out=path_out, optional=optional, block=block)


def wrap_git(
    command: str,
    *,
    out: str | None = None,
    workdir: str = "./",
    optional: bool = False,
    block: bool = False,
) -> StepInfo:
    """Create a step to run a git command.

    The provide the necessary input for StepUp, git is called through a wrapper script.
    This wrapper first infers the git root directory with `git rev-parse --show-toplevel`.
    It then amends the input dependencies with the following two files:

      - gitroot / '.git/HEAD'
      - gitroot / '.git/refs/branches/<branch>'

    As a result, the step will be rescheduled when the current commit id or branch changes.

    It is recommended to make gitroot / '.git/' and all of its contents static
    with recursive deferred glob as follows:

    ```python
    glob(".git/**", _defer=True)
    ```

    Parameters
    ----------
    command
        The git command to run, e.g. `git describe --tags`
        or `git log -n1 --pretty='format:%cs (%h)`.
    out
        An output file for the stdout of the git command.
    workdir
        The working directory where the git command must be executed.
    optional
        If `True`, the step is only executed when needed by other steps.
    block
        If `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    if not isinstance(command, str):
        raise TypeError("The git command must be a string.")
    if not (out is None or isinstance(out, str)):
        raise TypeError("The output of the git command must be a string or None.")

    action = "wrap-git"
    if out is not None:
        action += f" --out={shlex.quote(out)}"
    action += f" -- {command}"
    return step(
        action,
        out=out,
        workdir=workdir,
        optional=optional,
        block=block,
    )


def zip_inventory(
    path_inventory: str, path_zip: str, *, optional: bool = False, block: bool = False
) -> StepInfo:
    """Create a ZIP file with all files listed in a `inventory.txt` file + check digests before zip.

    Parameters
    ----------
    path_inventory
        A file created with the `make_inventory` API or with the command-line script
        `rr-make-inventory`.
    path_zip
        The output ZIP file
    optional
        If `True`, the step is only executed when needed by other steps.
    block
        If `True`, the step will always remain pending.

    Returns
    -------
    step_info
        Holds relevant information of the step, useful for defining follow-up steps.
    """
    return step(
        "zip-inventory ${inp} ${out}",
        inp=path_inventory,
        out=path_zip,
        optional=optional,
        block=block,
    )



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# © 2024–2025 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Put multiple pages per sheet using a fixed layout."""

import argparse
import sys

import fitz

from stepup.core.api import getenv

__all__ = ("nup_pdf",)


def main(argv: list[str] | None = None):
    """Main program."""
    args = parse_args(argv)
    if args.nrow is None:
        args.nrow = int(getenv("REPREP_NUP_NROW", "2"))
    if args.ncol is None:
        args.ncol = int(getenv("REPREP_NUP_NCOL", "2"))
    if args.margin is None:
        args.margin = float(getenv("REPREP_NUP_MARGIN", "10.0"))
    if args.page_format is None:
        args.page_format = getenv("REPREP_NUP_PAGE_FORMAT", "A4-L")
    nup_pdf(args.path_src, args.path_dst, args.nrow, args.ncol, args.margin, args.page_format)


def parse_args(argv: list[str] | None = None) -> argparse.Namespace:
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(
        prog="rr-nup-pdf", description="Put multiple pages per sheet using a fixed layout."
    )
    parser.add_argument("path_src", help="The source pdf to which notes should be added.")
    parser.add_argument("path_dst", help="The output pdf.")
    parser.add_argument(
        "-r",
        "--nrow",
        help="The number of rows. "
        "The default is ${REPREP_NUP_NROW} or 2 if the variable is not set.",
        type=int,
    )
    parser.add_argument(
        "-c",
        "--ncol",
        help="The number of columns. "
        "The default is ${REPREP_NUP_NCOL} or 2 if the variable is not set.",
        type=int,
    )
    parser.add_argument(
        "-m",
        "--margin",
        help="The margin in mm. "
        "The default is ${REPREP_NUP_MARGIN} or 10.0 if the variable is not set.",
        type=float,
    )
    parser.add_argument(
        "-p",
        "--page-format",
        help="The output page format. "
        "The default is ${REPREP_NUP_PAGE_FORMAT} or A4-L if the variable is not set.",
    )
    return parser.parse_args(argv)


def nup_pdf(
    path_src: str,
    path_dst: str,
    nrow: int,
    ncol: int,
    margin: float,
    page_format: str,
):
    """Put multiple pages in a single page, using a fixed layout.

    Parameters
    ----------
    path_src
        The source PDF filename.
    path_dst
        The destination PDF filename.
    nrow
        The number of rows in the layout.
    ncol
        The number of columns in the layout.
    margin
        The margin and (minimal) spacing between small pages in millimeter.
    page_format
        A string describing the output page size.
    """
    for path_pdf in path_src, path_dst:
        if not path_pdf.endswith(".pdf"):
            raise ValueError(
                f"All arguments must have a `.pdf` extension, got: {path_pdf}", file=sys.stderr
            )
    src = fitz.open(path_src)
    # See https://github.com/pymupdf/PyMuPDF/issues/3635
    src.scrub()
    dst = fitz.open()

    nup = nrow * ncol
    unit = 72 / 25.4
    # Convert distances in mm to points
    margin *= unit
    width, height = fitz.paper_size(page_format)

    # Spacing between two top-left corners of neighboring panels.
    xshift = (width - margin) / ncol
    yshift = (height - margin) / nrow

    # double loop adding all (small) pages to the destination PDF.
    for icoarse in range(0, len(src), nup):
        dst_page = dst.new_page(width=width, height=height)
        for ifine in range(icoarse, min(icoarse + nup, len(src))):
            ioffset = ifine - icoarse
            irow = ioffset // ncol
            icol = ioffset % ncol
            dst_page.show_pdf_page(
                fitz.Rect(
                    margin + xshift * icol,
                    margin + yshift * irow,
                    xshift * (icol + 1),
                    yshift * (irow + 1),
                ),
                src,
                ifine,
            )

    # Strip metadata for reproducibility and save
    dst.set_metadata({})
    dst.del_xml_metadata()
    dst.xref_set_key(-1, "ID", "null")
    dst.scrub()
    dst.save(path_dst, garbage=4, deflate=True, no_new_id=True)

    dst.close()
    src.close()


if __name__ == "__main__":
    main(sys.argv[1:])



# StepUp RepRep is the StepUp extension for Reproducible Reporting.
# © 2024–2025 Toon Verstraelen
#
# This file is part of StepUp RepRep.
#
# StepUp RepRep is free software;  you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# StepUp RepRep is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Utilities for testing with pytest."""

from path import Path

from stepup.core.pytest import run_example as run_example_core
from stepup.reprep.check_inventory import iter_inventory

__all__ = ("run_example",)


async def run_example(srcdir, tmpdir, overwrite_expected=False):
    """Run an example use case in a temporary directory and check the outputs.

    See stepup.core.pytest.run_example for details.
    """
    await run_example_core(srcdir, tmpdir, overwrite_expected=overwrite_expected)

    # Reproducibility check
    workdir = Path(tmpdir) / "example"
    for path_inventory in sorted(workdir.glob("reproducibility_*inventory.txt")):
        records = list(iter_inventory(path_inventory))
        sizes = {record.size for record in records}
        if len(sizes) != 1:
            raise AssertionError(f"Not all file sizes in {path_inventory} are the same.")
        digests = {record.digest for record in records}
        if len(digests) != 1:
            raise AssertionError(f"Not all file digests in {path_inventory} are the same.")
